import numpy as np
import pytest
import autoarray as aa


class TestRotations:
    def test__rotate_array__all_4_rotations_with_rotation_back(self):

        arr = np.array([[0.0, 1.0, 0.0], [1.0, 2.0, 0.0], [0.0, 0.0, 0.0]])

        arr_bl = aa.util.frame.rotate_array_from_roe_corner(
            array=arr, roe_corner=(1, 0)
        )

        assert arr_bl == pytest.approx(
            np.array([[0.0, 1.0, 0.0], [1.0, 2.0, 0.0], [0.0, 0.0, 0.0]]), 1.0e-4
        )

        arr_bl = aa.util.frame.rotate_array_from_roe_corner(
            array=arr_bl, roe_corner=(1, 0)
        )

        assert arr_bl == pytest.approx(np.array(arr), 1.0e-4)

        arr_br = aa.util.frame.rotate_array_from_roe_corner(
            array=arr, roe_corner=(1, 1)
        )

        assert arr_br == pytest.approx(
            np.array([[0.0, 1.0, 0.0], [0.0, 2.0, 1.0], [0.0, 0.0, 0.0]]), 1.0e-4
        )

        arr_br = aa.util.frame.rotate_array_from_roe_corner(
            array=arr_br, roe_corner=(1, 1)
        )

        assert arr_br == pytest.approx(np.array(arr), 1.0e-4)

        arr_tl = aa.util.frame.rotate_array_from_roe_corner(
            array=arr, roe_corner=(0, 0)
        )

        assert arr_tl == pytest.approx(
            np.array([[0.0, 0.0, 0.0], [1.0, 2.0, 0.0], [0.0, 1.0, 0.0]]), 1.0e-4
        )

        arr_tl = aa.util.frame.rotate_array_from_roe_corner(
            array=arr_tl, roe_corner=(0, 0)
        )

        assert arr_tl == pytest.approx(np.array(arr), 1.0e-4)

        arr_tr = aa.util.frame.rotate_array_from_roe_corner(
            array=arr, roe_corner=(0, 1)
        )

        assert arr_tr == pytest.approx(
            np.array([[0.0, 0.0, 0.0], [0.0, 2.0, 1.0], [0.0, 1.0, 0.0]]), 1.0e-4
        )

        arr_tr = aa.util.frame.rotate_array_from_roe_corner(
            array=arr_tr, roe_corner=(0, 1)
        )

        assert arr_tr == pytest.approx(np.array(arr), 1.0e-4)

    def test__rotate_region__all_4_rotations_with_rotation_back(self):

        region = (0, 2, 1, 3)

        shape_native = (8, 10)

        region_bl = aa.util.frame.rotate_region_from_roe_corner(
            region=region, shape_native=shape_native, roe_corner=(1, 0)
        )

        assert region_bl == (0, 2, 1, 3)

        region_bl = aa.util.frame.rotate_region_from_roe_corner(
            region=region_bl, shape_native=shape_native, roe_corner=(1, 0)
        )

        assert region_bl == (0, 2, 1, 3)

        region_br = aa.util.frame.rotate_region_from_roe_corner(
            region=region, shape_native=shape_native, roe_corner=(1, 1)
        )

        assert region_br == (0, 2, 7, 9)

        region_br = aa.util.frame.rotate_region_from_roe_corner(
            region=region_br, shape_native=shape_native, roe_corner=(1, 1)
        )

        assert region_br == (0, 2, 1, 3)

        region_tl = aa.util.frame.rotate_region_from_roe_corner(
            region=region, shape_native=shape_native, roe_corner=(0, 0)
        )

        assert region_tl == (6, 8, 1, 3)

        region_tl = aa.util.frame.rotate_region_from_roe_corner(
            region=region_tl, shape_native=shape_native, roe_corner=(0, 0)
        )

        assert region_tl == (0, 2, 1, 3)

        region_tr = aa.util.frame.rotate_region_from_roe_corner(
            region=region, shape_native=shape_native, roe_corner=(0, 1)
        )

        assert region_tr == (6, 8, 7, 9)

        region_tr = aa.util.frame.rotate_region_from_roe_corner(
            region=region_tr, shape_native=shape_native, roe_corner=(0, 1)
        )

        assert region_tr == (0, 2, 1, 3)

    # def test__rotate_ci_pattern__all_4_rotations_with_rotation_back(self):
    #
    #     pattern = ci_pattern.CIPatternUniform(
    #         regions=[(0, 1, 1, 2), (0, 2, 0, 2)], normalization=10.0
    #     )
    #
    #     shape_native = (2, 2)
    #
    #     pattern_bl = aa.util.frame.rotate_ci_pattern_from_roe_corner(
    #         ci_pattern=pattern, shape_native=shape_native, roe_corner=(1, 0)
    #     )
    #
    #     assert pattern_bl.regions == [(0, 1, 1, 2), (0, 2, 0, 2)]
    #     assert pattern_bl.normalization == 10.0
    #
    #     pattern_bl = aa.util.frame.rotate_ci_pattern_from_roe_corner(
    #         ci_pattern=pattern_bl, shape_native=shape_native, roe_corner=(1, 0)
    #     )
    #
    #     assert pattern_bl.regions == [(0, 1, 1, 2), (0, 2, 0, 2)]
    #
    #     pattern_br = aa.util.frame.rotate_ci_pattern_from_roe_corner(
    #         ci_pattern=pattern, shape_native=shape_native, roe_corner=(1, 1)
    #     )
    #
    #     assert pattern_br.regions == [(0, 1, 0, 1), (0, 2, 0, 2)]
    #
    #     pattern_br = aa.util.frame.rotate_ci_pattern_from_roe_corner(
    #         ci_pattern=pattern_br, shape_native=shape_native, roe_corner=(1, 1)
    #     )
    #
    #     assert pattern_br.regions == [(0, 1, 1, 2), (0, 2, 0, 2)]
    #
    #     pattern_tl = aa.util.frame.rotate_ci_pattern_from_roe_corner(
    #         ci_pattern=pattern, shape_native=shape_native, roe_corner=(0, 0)
    #     )
    #
    #     assert pattern_tl.regions == [(1, 2, 1, 2), (0, 2, 0, 2)]
    #
    #     pattern_tl = aa.util.frame.rotate_ci_pattern_from_roe_corner(
    #         ci_pattern=pattern_tl, shape_native=shape_native, roe_corner=(0, 0)
    #     )
    #
    #     assert pattern_tl.regions == [(0, 1, 1, 2), (0, 2, 0, 2)]
    #
    #     pattern_tr = aa.util.frame.rotate_ci_pattern_from_roe_corner(
    #         ci_pattern=pattern, shape_native=shape_native, roe_corner=(0, 1)
    #     )
    #
    #     assert pattern_tr.regions == [(1, 2, 0, 1), (0, 2, 0, 2)]
    #
    #     pattern_tr = aa.util.frame.rotate_ci_pattern_from_roe_corner(
    #         ci_pattern=pattern_tr, shape_native=shape_native, roe_corner=(0, 1)
    #     )
    #
    #     assert pattern_tr.regions == [(0, 1, 1, 2), (0, 2, 0, 2)]


class TestRegionAfterExtraction:
    def test__simple_test_cases(self):

        region = aa.util.frame.region_after_extraction(
            original_region=(2, 4, 2, 4), extraction_region=(0, 6, 0, 6)
        )

        assert region == (2, 4, 2, 4)

        region = aa.util.frame.region_after_extraction(
            original_region=(2, 4, 2, 4), extraction_region=(3, 5, 3, 5)
        )

        assert region == (0, 1, 0, 1)

        region = aa.util.frame.region_after_extraction(
            original_region=(2, 4, 2, 4), extraction_region=(2, 5, 2, 5)
        )

        assert region == (0, 2, 0, 2)

        region = aa.util.frame.region_after_extraction(
            original_region=(2, 4, 2, 4), extraction_region=(0, 3, 0, 3)
        )

        assert region == (2, 3, 2, 3)

    def test__regions_do_not_overlap__returns_none(self):

        region = aa.util.frame.region_after_extraction(
            original_region=(2, 4, 2, 4), extraction_region=(0, 6, 0, 1)
        )

        assert region == None

        region = aa.util.frame.region_after_extraction(
            original_region=(2, 4, 2, 4), extraction_region=(0, 1, 0, 6)
        )

        assert region == None

        region = aa.util.frame.region_after_extraction(
            original_region=(2, 4, 2, 4), extraction_region=(0, 1, 0, 1)
        )

        assert region == None

        region = aa.util.frame.region_after_extraction(
            original_region=None, extraction_region=(0, 6, 0, 1)
        )

        assert region == None


class Testx0x1AfterExtraction:
    def test__case_1__original_region_at_0__1d_extracted_region_is_fully_within_original_region(
        self,
    ):

        x0, x1 = aa.util.frame.x0x1_after_extraction(x0o=0, x1o=6, x0e=2, x1e=4)

        assert x0 == 0
        assert x1 == 2

        x0, x1 = aa.util.frame.x0x1_after_extraction(x0o=0, x1o=6, x0e=3, x1e=5)

        assert x0 == 0
        assert x1 == 2

        x0, x1 = aa.util.frame.x0x1_after_extraction(x0o=0, x1o=6, x0e=4, x1e=6)

        assert x0 == 0
        assert x1 == 2

        x0, x1 = aa.util.frame.x0x1_after_extraction(x0o=0, x1o=6, x0e=5, x1e=6)

        assert x0 == 0
        assert x1 == 1

        x0, x1 = aa.util.frame.x0x1_after_extraction(x0o=0, x1o=6, x0e=2, x1e=5)

        assert x0 == 0
        assert x1 == 3

    def test__case_2__original_region_offset_from_0__1d_extracted_region_is_fully_within_original_region(
        self,
    ):

        x0, x1 = aa.util.frame.x0x1_after_extraction(x0o=2, x1o=6, x0e=2, x1e=4)

        assert x0 == 0
        assert x1 == 2

        x0, x1 = aa.util.frame.x0x1_after_extraction(x0o=2, x1o=6, x0e=3, x1e=5)

        assert x0 == 0
        assert x1 == 2

        x0, x1 = aa.util.frame.x0x1_after_extraction(x0o=2, x1o=6, x0e=4, x1e=6)

        assert x0 == 0
        assert x1 == 2

        x0, x1 = aa.util.frame.x0x1_after_extraction(x0o=2, x1o=6, x0e=5, x1e=6)

        assert x0 == 0
        assert x1 == 1

        x0, x1 = aa.util.frame.x0x1_after_extraction(x0o=2, x1o=6, x0e=2, x1e=5)

        assert x0 == 0
        assert x1 == 3

    def test__case_3__original_region_offset_from_0__1d_extracted_region_partly_overlaps_to_left_original_region(
        self,
    ):

        x0, x1 = aa.util.frame.x0x1_after_extraction(x0o=2, x1o=6, x0e=1, x1e=3)

        assert x0 == 1
        assert x1 == 2

        x0, x1 = aa.util.frame.x0x1_after_extraction(x0o=2, x1o=6, x0e=1, x1e=4)

        assert x0 == 1
        assert x1 == 3

        x0, x1 = aa.util.frame.x0x1_after_extraction(x0o=2, x1o=6, x0e=0, x1e=3)

        assert x0 == 2
        assert x1 == 3

        x0, x1 = aa.util.frame.x0x1_after_extraction(x0o=2, x1o=6, x0e=0, x1e=5)

        assert x0 == 2
        assert x1 == 5

    def test__case_4__original_region_offset_from_0__1d_extracted_region_partly_overlaps_to_right_original_region(
        self,
    ):

        x0, x1 = aa.util.frame.x0x1_after_extraction(x0o=2, x1o=6, x0e=5, x1e=7)

        assert x0 == 0
        assert x1 == 1

        x0, x1 = aa.util.frame.x0x1_after_extraction(x0o=2, x1o=6, x0e=5, x1e=8)

        assert x0 == 0
        assert x1 == 1

        x0, x1 = aa.util.frame.x0x1_after_extraction(x0o=2, x1o=6, x0e=4, x1e=7)

        assert x0 == 0
        assert x1 == 2

        x0, x1 = aa.util.frame.x0x1_after_extraction(x0o=2, x1o=6, x0e=2, x1e=8)

        assert x0 == 0
        assert x1 == 4

    def test__case_5__extraction_region_over_full_original_region(self):

        x0, x1 = aa.util.frame.x0x1_after_extraction(x0o=2, x1o=6, x0e=0, x1e=8)

        assert x0 == 2
        assert x1 == 6

        x0, x1 = aa.util.frame.x0x1_after_extraction(x0o=2, x1o=6, x0e=0, x1e=7)

        assert x0 == 2
        assert x1 == 6

        x0, x1 = aa.util.frame.x0x1_after_extraction(x0o=2, x1o=6, x0e=1, x1e=8)

        assert x0 == 1
        assert x1 == 5

    def test__case_6__extraction_region_misses_original_region(self):

        x0, x1 = aa.util.frame.x0x1_after_extraction(x0o=2, x1o=6, x0e=7, x1e=8)

        assert x0 == None
        assert x1 == None

        x0, x1 = aa.util.frame.x0x1_after_extraction(x0o=2, x1o=6, x0e=6, x1e=8)

        assert x0 == None
        assert x1 == None

        x0, x1 = aa.util.frame.x0x1_after_extraction(x0o=2, x1o=6, x0e=0, x1e=1)

        assert x0 == None
        assert x1 == None

        x0, x1 = aa.util.frame.x0x1_after_extraction(x0o=2, x1o=6, x0e=0, x1e=2)

        assert x0 == None
        assert x1 == None



import numpy as np
import pytest

import autoarray as aa


def grid_to_pixel_pixels_via_nearest_neighbour(grid, pixel_centers):
    def compute_squared_separation(coordinate1, coordinate2):
        """
        Returns the squared separation of two grid (no square root for efficiency)"""
        return (coordinate1[0] - coordinate2[0]) ** 2 + (
            coordinate1[1] - coordinate2[1]
        ) ** 2

    image_pixels = grid.shape[0]

    image_to_pixelization = np.zeros((image_pixels,))

    for image_index, image_coordinate in enumerate(grid):
        distances = list(
            map(
                lambda centers: compute_squared_separation(image_coordinate, centers),
                pixel_centers,
            )
        )

        image_to_pixelization[image_index] = np.argmin(distances)

    return image_to_pixelization


class TestRectangularMapper:
    def test__sub_to_pix__various_grids__1_coordinate_per_square_pixel__in_centre_of_pixels(
        self,
    ):
        #   _ _ _
        #  I_I_I_I Boundaries for pixels x = 0 and y = 0  -1.0 to -(1/3)
        #  I_I_I_I Boundaries for pixels x = 1 and y = 1 - (1/3) to (1/3)
        #  I_I_I_I Boundaries for pixels x = 2 and y = 2 - (1/3)" to 1.0"

        grid = aa.Grid2D.manual_slim(
            [
                [1.0, -1.0],
                [1.0, 0.0],
                [1.0, 1.0],
                [0.0, -1.0],
                [0.0, 0.0],
                [0.0, 1.0],
                [-1.0, -1.0],
                [-1.0, 0.0],
                [-1.0, 1.0],
            ],
            pixel_scales=1.0,
            shape_native=(3, 3),
        )

        pixelization_grid = aa.Grid2DRectangular(
            grid=np.ones((2, 2)), shape_native=(3, 3), pixel_scales=(1.0, 1.0)
        )

        mapper = aa.Mapper(
            source_grid_slim=grid, source_pixelization_grid=pixelization_grid
        )

        assert (
            mapper.pixelization_index_for_sub_slim_index
            == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        ).all()

        print(mapper.all_sub_slim_indexes_for_pixelization_index)

        assert mapper.all_sub_slim_indexes_for_pixelization_index == [
            [0],
            [1],
            [2],
            [3],
            [4],
            [5],
            [6],
            [7],
            [8],
        ]

        #   _ _ _
        #  I_I_I_I Boundaries for pixels x = 0 and y = 0  -1.0 to -(1/3)
        #  I_I_I_I Boundaries for pixels x = 1 and y = 1 - (1/3) to (1/3)
        #  I_I_I_I Boundaries for pixels x = 2 and y = 2 - (1/3)" to 1.0"

        grid = aa.Grid2D.manual_slim(
            [
                [1.0, -1.0],
                [1.0, 0.0],
                [1.0, 1.0],
                [-0.32, -1.0],
                [-0.32, 0.32],
                [0.0, 1.0],
                [-0.34, -0.34],
                [-0.34, 0.325],
                [-1.0, 1.0],
            ],
            pixel_scales=1.0,
            shape_native=(3, 3),
        )

        pixelization_grid = aa.Grid2DRectangular.overlay_grid(
            shape_native=(3, 3), grid=grid
        )

        mapper = aa.Mapper(
            source_grid_slim=grid, source_pixelization_grid=pixelization_grid
        )

        assert (
            mapper.pixelization_index_for_sub_slim_index
            == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        ).all()

        assert mapper.all_sub_slim_indexes_for_pixelization_index == [
            [0],
            [1],
            [2],
            [3],
            [4],
            [5],
            [6],
            [7],
            [8],
        ]

    def test__sub_to_pix__3x3_grid_of_pixel_grid__add_multiple_grid_to_1_pixel(self):
        #                  _ _ _
        # -1.0 to -(1/3)  I_I_I_I
        # -(1/3) to (1/3) I_I_I_I
        #  (1/3) to 1.0   I_I_I_I

        grid = aa.Grid2D.manual_slim(
            [
                [1.0, -1.0],
                [0.0, 0.0],
                [1.0, 1.0],
                [0.0, 0.0],
                [0.0, 0.0],
                [0.0, 0.0],
                [-1.0, -1.0],
                [0.0, 0.0],
                [-1.0, 1.0],
            ],
            pixel_scales=1.0,
            shape_native=(3, 3),
        )

        pixelization_grid = aa.Grid2DRectangular.overlay_grid(
            shape_native=(3, 3), grid=grid
        )

        mapper = aa.Mapper(
            source_grid_slim=grid, source_pixelization_grid=pixelization_grid
        )

        assert (
            mapper.pixelization_index_for_sub_slim_index
            == np.array([0, 4, 2, 4, 4, 4, 6, 4, 8])
        ).all()

        assert mapper.all_sub_slim_indexes_for_pixelization_index == [
            [0],
            [],
            [2],
            [],
            [1, 3, 4, 5, 7],
            [],
            [6],
            [],
            [8],
        ]

    def test__sub_to_pix__various_grids__1_coordinate_in_each_pixel(self):
        #   _ _ _
        #  I_I_I_I
        #  I_I_I_I
        #  I_I_I_I
        #  I_I_I_I

        # Boundaries for column pixel 0 -1.0 to -(1/3)
        # Boundaries for column pixel 1 -(1/3) to (1/3)
        # Boundaries for column pixel 2  (1/3) to 1.0

        # Bounadries for row pixel 0 -1.0 to -0.5
        # Bounadries for row pixel 1 -0.5 to 0.0
        # Bounadries for row pixel 2  0.0 to 0.5
        # Bounadries for row pixel 3  0.5 to 1.0

        grid = aa.Grid2D.manual_slim(
            [
                [1.0, -1.0],
                [1.0, 0.0],
                [1.0, 1.0],
                [0.5, -1.0],
                [-0.5, 1.0],
                [-1.0, 1.0],
            ],
            pixel_scales=1.0,
            shape_native=(3, 2),
        )

        pixelization_grid = aa.Grid2DRectangular.overlay_grid(
            shape_native=(4, 3), grid=grid
        )

        mapper = aa.Mapper(
            source_grid_slim=grid, source_pixelization_grid=pixelization_grid
        )

        assert (
            mapper.pixelization_index_for_sub_slim_index
            == np.array([0, 1, 2, 3, 8, 11])
        ).all()

        assert mapper.all_sub_slim_indexes_for_pixelization_index == [
            [0],
            [1],
            [2],
            [3],
            [],
            [],
            [],
            [],
            [4],
            [],
            [],
            [5],
        ]

        #   _ _ _ _
        #  I_I_I_I_I
        #  I_I_I_I_I
        #  I_I_I_I_I

        # Boundaries for row pixel 0 -1.0 to -(1/3)
        # Boundaries for row pixel 1 -(1/3) to (1/3)
        # Boundaries for row pixel 2  (1/3) to 1.0

        # Bounadries for column pixel 0 -1.0 to -0.5
        # Bounadries for column pixel 1 -0.5 to 0.0
        # Bounadries for column pixel 2  0.0 to 0.5
        # Bounadries for column pixel 3  0.5 to 1.0

        grid = aa.Grid2D.manual_slim(
            [
                [1.0, -1.0],
                [1.0, -0.49],
                [1.0, 0.01],
                [0.32, 0.01],
                [-0.34, -0.01],
                [-1.0, 1.0],
            ],
            pixel_scales=1.0,
            shape_native=(2, 3),
        )

        pixelization_grid = aa.Grid2DRectangular.overlay_grid(
            shape_native=(3, 4), grid=grid
        )

        mapper = aa.Mapper(
            source_grid_slim=grid, source_pixelization_grid=pixelization_grid
        )

        assert (
            mapper.pixelization_index_for_sub_slim_index
            == np.array([0, 1, 2, 6, 9, 11])
        ).all()

        assert mapper.all_sub_slim_indexes_for_pixelization_index == [
            [0],
            [1],
            [2],
            [],
            [],
            [],
            [3],
            [],
            [],
            [4],
            [],
            [5],
        ]

    def test__sub_to_pix__3x3_grid__change_scaledond_dimensions_size__grid_adapts_accordingly(
        self,
    ):
        #   _ _ _
        #  I_I_I_I Boundaries for pixels x = 0 and y = 0  -1.5 to -0.5
        #  I_I_I_I Boundaries for pixels x = 1 and y = 1 -0.5 to 0.5
        #  I_I_I_I Boundaries for pixels x = 2 and y = 2  0.5 to 1.5

        grid = aa.Grid2D.manual_slim(
            [[1.5, -1.5], [1.0, 0.0], [1.0, 0.6], [-1.4, 0.0], [-1.5, 1.5]],
            pixel_scales=1.0,
            shape_native=(5, 1),
        )

        pixelization_grid = aa.Grid2DRectangular.overlay_grid(
            shape_native=(3, 3), grid=grid
        )

        mapper = aa.Mapper(
            source_grid_slim=grid, source_pixelization_grid=pixelization_grid
        )

        assert (
            mapper.pixelization_index_for_sub_slim_index == np.array([0, 1, 2, 7, 8])
        ).all()

        assert mapper.all_sub_slim_indexes_for_pixelization_index == [
            [0],
            [1],
            [2],
            [],
            [],
            [],
            [],
            [3],
            [4],
        ]

    def test__sub_to_pix__various_grids__change_scaledond_dimensions__not_symmetric(
        self,
    ):
        #   _ _ _
        #  I_I_I_I Boundaries for pixels x = 0 and y = 0  -1.5 to -0.5
        #  I_I_I_I Boundaries for pixels x = 1 and y = 1 -0.5 to 0.5
        #  I_I_I_I Boundaries for pixels x = 2 and y = 2  0.5 to 1.5

        grid = aa.Grid2D.manual_slim(
            [[1.0, -1.5], [1.0, -0.49], [0.32, -1.5], [0.32, 0.51], [-1.0, 1.5]],
            pixel_scales=1.0,
            shape_native=(5, 1),
        )

        pixelization_grid = aa.Grid2DRectangular.overlay_grid(
            shape_native=(3, 3), grid=grid
        )

        mapper = aa.Mapper(
            source_grid_slim=grid, source_pixelization_grid=pixelization_grid
        )

        assert (
            mapper.pixelization_index_for_sub_slim_index == np.array([0, 1, 3, 5, 8])
        ).all()

        assert mapper.all_sub_slim_indexes_for_pixelization_index == [
            [0],
            [1],
            [],
            [2],
            [],
            [3],
            [],
            [],
            [4],
        ]

        #   _ _ _
        #  I_I_I_I
        #  I_I_I_I
        #  I_I_I_I
        #  I_I_I_I

        grid = aa.Grid2D.manual_slim(
            [[1.0, -1.5], [1.0, -0.49], [0.49, -1.5], [-0.6, 0.0], [-1.0, 1.5]],
            pixel_scales=1.0,
            shape_native=(5, 1),
        )

        pixelization_grid = aa.Grid2DRectangular.overlay_grid(
            shape_native=(4, 3), grid=grid
        )

        mapper = aa.Mapper(
            source_grid_slim=grid, source_pixelization_grid=pixelization_grid
        )

        assert (
            mapper.pixelization_index_for_sub_slim_index == np.array([0, 1, 3, 10, 11])
        ).all()

        assert mapper.all_sub_slim_indexes_for_pixelization_index == [
            [0],
            [1],
            [],
            [2],
            [],
            [],
            [],
            [],
            [],
            [],
            [3],
            [4],
        ]

        #   _ _ _ _
        #  I_I_I_I_I
        #  I_I_I_I_I
        #  I_I_I_I_I

        grid = aa.Grid2D.manual_slim(
            [[1.0, -1.5], [1.0, -0.49], [0.32, -1.5], [-0.34, 0.49], [-1.0, 1.5]],
            pixel_scales=1.0,
            shape_native=(5, 1),
        )

        pixelization_grid = aa.Grid2DRectangular.overlay_grid(
            shape_native=(3, 4), grid=grid
        )

        mapper = aa.Mapper(
            source_grid_slim=grid, source_pixelization_grid=pixelization_grid
        )

        assert (
            mapper.pixelization_index_for_sub_slim_index == np.array([0, 1, 4, 10, 11])
        ).all()

        assert mapper.all_sub_slim_indexes_for_pixelization_index == [
            [0],
            [1],
            [],
            [],
            [2],
            [],
            [],
            [],
            [],
            [],
            [3],
            [4],
        ]

    def test__sub_to_pix__different_image_and_sub_grids(self):
        #                  _ _ _
        # -1.0 to -(1/3)  I_I_I_I
        # -(1/3) to (1/3) I_I_I_I
        #  (1/3) to 1.0   I_I_I_I

        grid = aa.Grid2D.manual_slim(
            [
                [1.0, -1.0],
                [1.0, 0.0],
                [1.0, 1.0],
                [0.0, -1.0],
                [0.0, 0.0],
                [0.0, 1.0],
                [-1.0, -1.0],
                [-1.0, 0.0],
                [-1.0, 1.0],
            ],
            pixel_scales=1.0,
            shape_native=(3, 3),
        )

        pixelization_grid = aa.Grid2DRectangular.overlay_grid(
            shape_native=(3, 3), grid=grid
        )

        mapper = aa.Mapper(
            source_grid_slim=grid, source_pixelization_grid=pixelization_grid
        )

        assert (
            mapper.pixelization_index_for_sub_slim_index
            == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        ).all()

        assert mapper.all_sub_slim_indexes_for_pixelization_index == [
            [0],
            [1],
            [2],
            [3],
            [4],
            [5],
            [6],
            [7],
            [8],
        ]

    def test__sub_to_pix__3x3_grid_of_pixel_grid___shift_coordinates_to_new_centre__centre_adjusts_based_on_grid(
        self,
    ):
        #   _ _ _
        #  I_I_I_I Boundaries for pixels x = 0 and y = 0  -1.0 to -(1/3)
        #  I_I_I_I Boundaries for pixels x = 1 and y = 1 - (1/3) to (1/3)
        #  I_I_I_I Boundaries for pixels x = 2 and y = 2 - (1/3)" to 1.0"

        grid = aa.Grid2D.manual_slim(
            [
                [2.0, 0.0],
                [2.0, 1.0],
                [2.0, 2.0],
                [1.0, 0.0],
                [1.0, 1.0],
                [1.0, 2.0],
                [0.0, 0.0],
                [0.0, 1.0],
                [0.0, 2.0],
            ],
            pixel_scales=1.0,
            shape_native=(3, 3),
        )

        pixelization_grid = aa.Grid2DRectangular.overlay_grid(
            shape_native=(3, 3), grid=grid
        )

        mapper = aa.Mapper(
            source_grid_slim=grid, source_pixelization_grid=pixelization_grid
        )

        assert (
            mapper.pixelization_index_for_sub_slim_index
            == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        ).all()

        assert mapper.all_sub_slim_indexes_for_pixelization_index == [
            [0],
            [1],
            [2],
            [3],
            [4],
            [5],
            [6],
            [7],
            [8],
        ]

    def test__sub_to_pix__4x3_grid__non_symmetric_centre_shift(self):
        #   _ _ _
        #  I_I_I_I
        #  I_I_I_I
        #  I_I_I_I
        #  I_I_I_I

        grid = aa.Grid2D.manual_slim(
            [[3.0, -0.5], [3.0, 0.51], [2.49, -0.5], [1.4, 1.0], [1.0, 2.5]],
            pixel_scales=1.0,
            shape_native=(5, 1),
        )

        pixelization_grid = aa.Grid2DRectangular.overlay_grid(
            shape_native=(4, 3), grid=grid
        )

        mapper = aa.Mapper(
            source_grid_slim=grid, source_pixelization_grid=pixelization_grid
        )

        assert (
            mapper.pixelization_index_for_sub_slim_index == np.array([0, 1, 3, 10, 11])
        ).all()
        assert mapper.all_sub_slim_indexes_for_pixelization_index == [
            [0],
            [1],
            [],
            [2],
            [],
            [],
            [],
            [],
            [],
            [],
            [3],
            [4],
        ]

    def test__reconstructed_pixelization__3x3_pixelization__solution_vector_ascending(
        self,
    ):
        grid = aa.Grid2D.manual_slim(
            [
                [2.0, -1.0],
                [2.0, 0.0],
                [2.0, 1.0],
                [0.0, -1.0],
                [0.0, 0.0],
                [0.0, 1.0],
                [-2.0, -1.0],
                [-2.0, 0.0],
                [-2.0, 1.0],
            ],
            pixel_scales=1.0,
            shape_native=(3, 3),
        )

        pixelization_grid = aa.Grid2DRectangular.overlay_grid(
            shape_native=(3, 3), grid=grid
        )

        mapper = aa.Mapper(
            source_grid_slim=grid, source_pixelization_grid=pixelization_grid
        )

        recon_pix = mapper.reconstruction_from(
            solution_vector=np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0])
        )

        assert (
            recon_pix.native
            == np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]])
        ).all()
        assert recon_pix.pixel_scales == pytest.approx((4.0 / 3.0, 2.0 / 3.0), 1e-2)
        assert recon_pix.origin == (0.0, 0.0)

    def test__reconstructed_pixelization__compare_to_imaging_util(self):

        grid = aa.Grid2D.manual_slim(
            [[1.0, 1.0], [1.0, 1.0], [1.0, 1.0], [1.0, 1.0]],
            pixel_scales=1.0,
            shape_native=(2, 2),
        )

        pixelization_grid = aa.Grid2DRectangular.overlay_grid(
            shape_native=(4, 3), grid=grid
        )

        mapper = aa.Mapper(
            source_grid_slim=grid, source_pixelization_grid=pixelization_grid
        )

        solution = np.array(
            [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 1.0, 2.0, 3.0]
        )
        recon_pix = mapper.reconstruction_from(solution_vector=solution)
        recon_pix_util = aa.util.array.sub_array_2d_from(
            sub_array_2d_slim=solution,
            mask_2d=np.full(fill_value=False, shape=(4, 3)),
            sub_size=1,
        )
        assert (recon_pix.native == recon_pix_util).all()
        assert recon_pix.shape_native == (4, 3)

        pixelization_grid = aa.Grid2DRectangular.overlay_grid(
            shape_native=(3, 4), grid=grid
        )

        mapper = aa.Mapper(
            source_grid_slim=grid, source_pixelization_grid=pixelization_grid
        )

        solution = np.array(
            [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 1.0, 2.0, 3.0]
        )
        recon_pix = mapper.reconstruction_from(solution_vector=solution)
        recon_pix_util = aa.util.array.sub_array_2d_from(
            sub_array_2d_slim=solution,
            mask_2d=np.full(fill_value=False, shape=(3, 4)),
            sub_size=1,
        )
        assert (recon_pix.native == recon_pix_util).all()
        assert recon_pix.shape_native == (3, 4)

    def test__pixel_signals__compare_to_mapper_util(self, grid_7x7, image_7x7):
        pixelization_grid = aa.Grid2DRectangular.overlay_grid(
            shape_native=(3, 3), grid=grid_7x7
        )

        mapper = aa.Mapper(
            source_grid_slim=grid_7x7,
            source_pixelization_grid=pixelization_grid,
            hyper_data=image_7x7,
        )

        pixel_signals = mapper.pixel_signals_from_signal_scale(signal_scale=2.0)

        pixel_signals_util = aa.util.mapper.adaptive_pixel_signals_from(
            pixels=9,
            signal_scale=2.0,
            pixelization_index_for_sub_slim_index=mapper.pixelization_index_for_sub_slim_index,
            slim_index_for_sub_slim_index=grid_7x7.mask._slim_index_for_sub_slim_index,
            hyper_image=image_7x7,
        )

        assert (pixel_signals == pixel_signals_util).all()

    def test__image_from_source__different_types_of_lists_input(self, sub_grid_7x7):

        rectangular_pixelization_grid = aa.Grid2DRectangular.overlay_grid(
            grid=sub_grid_7x7, shape_native=(3, 3)
        )
        rectangular_mapper = aa.Mapper(
            source_grid_slim=sub_grid_7x7,
            source_pixelization_grid=rectangular_pixelization_grid,
        )

        full_indexes = rectangular_mapper.slim_indexes_from_pixelization_indexes(
            pixelization_indexes=[0, 1]
        )

        assert full_indexes == [0, 1, 2, 3, 4, 5, 6, 7]

        full_indexes = rectangular_mapper.slim_indexes_from_pixelization_indexes(
            pixelization_indexes=[[0], [4]]
        )

        assert full_indexes == [[0, 1, 2, 3], [16, 17, 18, 19]]


class TestVoronoiMapper:
    def test__grid_to_pixel_pixels_via_nearest_neighbour__case1__correct_pairs(self):
        pixel_centers = np.array([[1.0, 1.0], [-1.0, 1.0], [-1.0, -1.0], [1.0, -1.0]])
        grid = aa.Grid2D.manual_slim(
            [[1.1, 1.1], [-1.1, 1.1], [-1.1, -1.1], [1.1, -1.1]],
            shape_native=(2, 2),
            pixel_scales=1.0,
        )

        sub_to_pix = grid_to_pixel_pixels_via_nearest_neighbour(grid, pixel_centers)

        assert sub_to_pix[0] == 0
        assert sub_to_pix[1] == 1
        assert sub_to_pix[2] == 2
        assert sub_to_pix[3] == 3

    def test__grid_to_pixel_pixels_via_nearest_neighbour___case2__correct_pairs(self):
        pixel_centers = np.array([[1.0, 1.0], [-1.0, 1.0], [-1.0, -1.0], [1.0, -1.0]])
        grid = aa.Grid2D.manual_slim(
            [
                [1.1, 1.1],
                [-1.1, 1.1],
                [-1.1, -1.1],
                [1.1, -1.1],
                [0.9, -0.9],
                [-0.9, -0.9],
                [-0.9, 0.9],
                [0.9, 0.9],
            ],
            shape_native=(3, 3),
            pixel_scales=0.1,
        )

        sub_to_pix = grid_to_pixel_pixels_via_nearest_neighbour(grid, pixel_centers)

        assert sub_to_pix[0] == 0
        assert sub_to_pix[1] == 1
        assert sub_to_pix[2] == 2
        assert sub_to_pix[3] == 3
        assert sub_to_pix[4] == 3
        assert sub_to_pix[5] == 2
        assert sub_to_pix[6] == 1
        assert sub_to_pix[7] == 0

    def test__grid_to_pixel_pixels_via_nearest_neighbour___case3__correct_pairs(self):
        pixel_centers = np.array(
            [[1.0, 1.0], [-1.0, 1.0], [-1.0, -1.0], [1.0, -1.0], [0.0, 0.0], [2.0, 2.0]]
        )
        grid = aa.Grid2D.manual_slim(
            [
                [0.1, 0.1],
                [-0.1, -0.1],
                [0.49, 0.49],
                [0.51, 0.51],
                [1.01, 1.01],
                [1.51, 1.51],
            ],
            shape_native=(3, 2),
            pixel_scales=1.0,
        )

        sub_to_pix = grid_to_pixel_pixels_via_nearest_neighbour(grid, pixel_centers)

        assert sub_to_pix[0] == 4
        assert sub_to_pix[1] == 4
        assert sub_to_pix[2] == 4
        assert sub_to_pix[3] == 0
        assert sub_to_pix[4] == 0
        assert sub_to_pix[5] == 5

    def test__sub_to_pix_of_mapper_matches_nearest_neighbor_calculation(self, grid_7x7):
        pixelization_grid = aa.Grid2D.manual_slim(
            [[0.1, 0.1], [1.1, 0.1], [2.1, 0.1], [0.1, 1.1], [1.1, 1.1], [2.1, 1.1]],
            shape_native=(3, 2),
            pixel_scales=1.0,
        )

        sub_to_pix_nearest_neighbour = grid_to_pixel_pixels_via_nearest_neighbour(
            grid_7x7, pixelization_grid
        )

        nearest_pixelization_index_for_slim_index = np.array(
            [0, 0, 1, 0, 0, 1, 2, 2, 3]
        )

        pixelization_grid = aa.Grid2DVoronoi(
            grid=pixelization_grid,
            nearest_pixelization_index_for_slim_index=nearest_pixelization_index_for_slim_index,
        )

        mapper = aa.Mapper(
            source_grid_slim=grid_7x7, source_pixelization_grid=pixelization_grid
        )

        assert (
            mapper.pixelization_index_for_sub_slim_index == sub_to_pix_nearest_neighbour
        ).all()

    def test__pixel_scales___for_voronoi_mapper(self, grid_7x7, image_7x7):
        pixelization_grid = aa.Grid2D.manual_slim(
            [[0.1, 0.1], [1.1, 0.1], [2.1, 0.1], [0.1, 1.1], [1.1, 1.1], [2.1, 1.1]],
            shape_native=(3, 2),
            pixel_scales=1.0,
        )

        nearest_pixelization_index_for_slim_index = np.array(
            [0, 0, 1, 0, 0, 1, 2, 2, 3]
        )

        pixelization_grid = aa.Grid2DVoronoi(
            grid=pixelization_grid,
            nearest_pixelization_index_for_slim_index=nearest_pixelization_index_for_slim_index,
        )

        mapper = aa.Mapper(
            source_grid_slim=grid_7x7,
            source_pixelization_grid=pixelization_grid,
            hyper_data=image_7x7,
        )

        pixel_signals = mapper.pixel_signals_from_signal_scale(signal_scale=2.0)

        pixel_signals_util = aa.util.mapper.adaptive_pixel_signals_from(
            pixels=6,
            signal_scale=2.0,
            pixelization_index_for_sub_slim_index=mapper.pixelization_index_for_sub_slim_index,
            slim_index_for_sub_slim_index=grid_7x7.mask._slim_index_for_sub_slim_index,
            hyper_image=image_7x7,
        )

        assert (pixel_signals == pixel_signals_util).all()



import os
from os import path
import shutil
import numpy as np
import pytest

import autoarray as aa
from autoarray.structures import grids

test_grid_dir = path.join(
    "{}".format(path.dirname(path.realpath(__file__))), "files", "grid"
)


class TestAbstractGrid2DIrregular:
    def test__structure_from_result__maps_numpy_array_to__auto_array_or_grid(self):

        grid = aa.Grid2DIrregular(grid=[(1.0, -1.0), (1.0, 1.0)])

        result = grid.structure_from_result(result=np.array([1.0, 2.0]))

        assert isinstance(result, aa.ValuesIrregularGrouped)
        assert result.in_list == [1.0, 2.0]

        result = grid.structure_from_result(result=np.array([[1.0, 1.0], [2.0, 2.0]]))

        assert isinstance(result, aa.Grid2DIrregular)
        assert result.in_list == [(1.0, 1.0), (2.0, 2.0)]

        grid = aa.Grid2DIrregularGrouped(grid=[(1.0, -1.0), (1.0, 1.0)])

        result = grid.structure_from_result(result=np.array([1.0, 2.0]))

        assert isinstance(result, aa.ValuesIrregularGrouped)
        assert result.in_grouped_list == [[1.0, 2.0]]

        result = grid.structure_from_result(result=np.array([[1.0, 1.0], [2.0, 2.0]]))

        assert isinstance(result, aa.Grid2DIrregularGrouped)
        assert result.in_grouped_list == [[(1.0, 1.0), (2.0, 2.0)]]

    def test__structure_list_from_result_list__maps_list_to_auto_arrays_or_grids(self):

        grid = aa.Grid2DIrregular(grid=[(1.0, -1.0), (1.0, 1.0)])

        result = grid.structure_from_result(result=[np.array([1.0, 2.0])])

        assert isinstance(result[0], aa.ValuesIrregularGrouped)
        assert result[0].in_list == [1.0, 2.0]

        result = grid.structure_from_result(result=[np.array([[1.0, 1.0], [2.0, 2.0]])])

        assert isinstance(result[0], aa.Grid2DIrregular)
        assert result[0].in_list == [(1.0, 1.0), (2.0, 2.0)]

        grid = aa.Grid2DIrregularGrouped(grid=[(1.0, -1.0), (1.0, 1.0)])

        result = grid.structure_from_result(result=[np.array([1.0, 2.0])])

        assert isinstance(result[0], aa.ValuesIrregularGrouped)
        assert result[0].in_grouped_list == [[1.0, 2.0]]

        result = grid.structure_from_result(result=[np.array([[1.0, 1.0], [2.0, 2.0]])])

        assert isinstance(result[0], aa.Grid2DIrregularGrouped)
        assert result[0].in_grouped_list == [[(1.0, 1.0), (2.0, 2.0)]]


class TestGrid2DIrregular:
    def test__input_as_list_or_list_of_other_types__all_convert_correctly(self):

        # Input tuples

        grid = aa.Grid2DIrregular(grid=[(1.0, -1.0), (1.0, 1.0)])

        assert type(grid) == grids.Grid2DIrregular
        assert (grid == np.array([[[1.0, -1.0], [1.0, 1.0]]])).all()
        assert grid.in_list == [(1.0, -1.0), (1.0, 1.0)]

        # Input np array

        grid = aa.Grid2DIrregular(grid=[np.array([1.0, -1.0]), np.array([1.0, 1.0])])

        assert type(grid) == grids.Grid2DIrregular
        assert (grid == np.array([[[1.0, -1.0], [1.0, 1.0]]])).all()
        assert grid.in_list == [(1.0, -1.0), (1.0, 1.0)]

        # Input list

        grid = aa.Grid2DIrregular(grid=[[1.0, -1.0], [1.0, 1.0]])

        assert type(grid) == grids.Grid2DIrregular
        assert (grid == np.array([[[1.0, -1.0], [1.0, 1.0]]])).all()
        assert grid.in_list == [(1.0, -1.0), (1.0, 1.0)]

    def test__values_from_array_slim(self):

        grid = aa.Grid2DIrregular(grid=[(1.0, 1.0), (2.0, 2.0)])

        values_from_1d = grid.values_from_array_slim(array_slim=np.array([1.0, 2.0]))

        assert isinstance(values_from_1d, aa.ValuesIrregularGrouped)
        assert values_from_1d.in_list == [1.0, 2.0]

        grid = aa.Grid2DIrregular(grid=[(1.0, 1.0), (2.0, 2.0), (3.0, 3.0)])

        values_from_1d = grid.values_from_array_slim(
            array_slim=np.array([1.0, 2.0, 3.0])
        )

        assert isinstance(values_from_1d, aa.ValuesIrregularGrouped)
        assert values_from_1d.in_list == [1.0, 2.0, 3.0]

    def test__grid_from_grid_slim(self):

        grid = aa.Grid2DIrregular(grid=[(1.0, 1.0), (2.0, 2.0)])

        grid_from_1d = grid.grid_from_grid_slim(
            grid_slim=np.array([[1.0, 1.0], [2.0, 2.0]])
        )

        assert type(grid_from_1d) == grids.Grid2DIrregular
        assert grid_from_1d.in_list == [(1.0, 1.0), (2.0, 2.0)]

        grid = aa.Grid2DIrregular(grid=[(1.0, 1.0), (2.0, 2.0), (3.0, 3.0)])

        grid_from_1d = grid.grid_from_grid_slim(
            grid_slim=np.array([[1.0, 1.0], [2.0, 2.0], [3.0, 3.0]])
        )

        assert type(grid_from_1d) == grids.Grid2DIrregular
        assert grid_from_1d.in_list == [(1.0, 1.0), (2.0, 2.0), (3.0, 3.0)]

    def test__grid_from_deflection_grid(self):

        grid = aa.Grid2DIrregular(grid=[(1.0, 1.0), (2.0, 2.0)])

        grid = grid.grid_from_deflection_grid(
            deflection_grid=np.array([[1.0, 0.0], [1.0, 1.0]])
        )

        assert type(grid) == grids.Grid2DIrregular
        assert grid.in_list == [(0.0, 1.0), (1.0, 1.0)]

    def test__furthest_distances_from_other_coordinates(self):

        grid = aa.Grid2DIrregular(grid=[(0.0, 0.0), (0.0, 1.0)])

        assert grid.furthest_distances_from_other_coordinates.in_grouped_list == [
            [1.0, 1.0]
        ]

        grid = aa.Grid2DIrregular(grid=[(2.0, 4.0), (3.0, 6.0)])

        assert grid.furthest_distances_from_other_coordinates.in_grouped_list == [
            [np.sqrt(5), np.sqrt(5)]
        ]

        grid = aa.Grid2DIrregular(grid=[(0.0, 0.0), (0.0, 1.0), (0.0, 3.0)])

        assert grid.furthest_distances_from_other_coordinates.in_grouped_list == [
            [3.0, 2.0, 3.0]
        ]

    def test__grid_of_closest_from_grid(self):

        grid = aa.Grid2DIrregular(grid=[(0.0, 0.0), (0.0, 1.0)])

        grid_of_closest = grid.grid_of_closest_from_grid_pair(
            grid_pair=np.array([[0.0, 0.1]])
        )

        assert (grid_of_closest == np.array([[0.0, 0.0]])).all()

        grid_of_closest = grid.grid_of_closest_from_grid_pair(
            grid_pair=np.array([[0.0, 0.1], [0.0, 0.2], [0.0, 0.3]])
        )

        assert (grid_of_closest == np.array([[0.0, 0.0], [0.0, 0.0], [0.0, 0.0]])).all()

        grid_of_closest = grid.grid_of_closest_from_grid_pair(
            grid_pair=np.array([[0.0, 0.1], [0.0, 0.2], [0.0, 0.9], [0.0, -0.1]])
        )

        assert (
            grid_of_closest
            == np.array([[0.0, 0.0], [0.0, 0.0], [0.0, 1.0], [0.0, 0.0]])
        ).all()


class TestGrid2DIrregularGrouped:
    def test__indexes_give_entries_where_list_begin_and_end(self):

        grid = aa.Grid2DIrregularGrouped(grid=[[(0.0, 0.0)]])

        assert grid.lower_indexes == [0]
        assert grid.upper_indexes == [1]

        grid = aa.Grid2DIrregularGrouped(grid=[[(0.0, 0.0), (0.0, 0.0)]])

        assert grid.lower_indexes == [0]
        assert grid.upper_indexes == [2]

        grid = aa.Grid2DIrregularGrouped(grid=[[(0.0, 0.0), (0.0, 0.0)], [(0.0, 0.0)]])

        assert grid.lower_indexes == [0, 2]
        assert grid.upper_indexes == [2, 3]

        grid = aa.Grid2DIrregularGrouped(
            grid=[
                [(0.0, 0.0), (0.0, 0.0)],
                [(0.0, 0.0), (0.0, 0.0), (0.0, 0.0)],
                [(0.0, 0.0), (0.0, 0.0)],
                [(0.0, 0.0)],
            ]
        )

        assert grid.lower_indexes == [0, 2, 5, 7]
        assert grid.upper_indexes == [2, 5, 7, 8]

    def test__input_as_list_or_list_of_other_types__all_convert_correctly(self):

        # Input tuples

        grid = aa.Grid2DIrregularGrouped(grid=[(1.0, -1.0), (1.0, 1.0)])

        assert type(grid) == grids.Grid2DIrregularGrouped
        assert (grid == np.array([[[1.0, -1.0], [1.0, 1.0]]])).all()
        assert grid.in_grouped_list == [[(1.0, -1.0), (1.0, 1.0)]]
        assert grid.in_list == [(1.0, -1.0), (1.0, 1.0)]

        grid = aa.Grid2DIrregularGrouped(grid=[[(1.0, -1.0), (1.0, 1.0)]])

        assert type(grid) == grids.Grid2DIrregularGrouped
        assert (grid == np.array([[[1.0, -1.0], [1.0, 1.0]]])).all()
        assert grid.in_grouped_list == [[(1.0, -1.0), (1.0, 1.0)]]
        assert grid.in_list == [(1.0, -1.0), (1.0, 1.0)]

        # Input np array

        grid = aa.Grid2DIrregularGrouped(
            grid=[np.array([1.0, -1.0]), np.array([1.0, 1.0])]
        )

        assert type(grid) == grids.Grid2DIrregularGrouped
        assert (grid == np.array([[[1.0, -1.0], [1.0, 1.0]]])).all()
        assert grid.in_grouped_list == [[(1.0, -1.0), (1.0, 1.0)]]
        assert grid.in_list == [(1.0, -1.0), (1.0, 1.0)]

        grid = aa.Grid2DIrregularGrouped(
            grid=[[np.array([1.0, -1.0]), np.array([1.0, 1.0])]]
        )

        assert type(grid) == grids.Grid2DIrregularGrouped
        assert (grid == np.array([[[1.0, -1.0], [1.0, 1.0]]])).all()
        assert grid.in_grouped_list == [[(1.0, -1.0), (1.0, 1.0)]]
        assert grid.in_list == [(1.0, -1.0), (1.0, 1.0)]

        grid = aa.Grid2DIrregularGrouped(grid=[np.array([[1.0, -1.0], [1.0, 1.0]])])

        assert type(grid) == grids.Grid2DIrregularGrouped
        assert (grid == np.array([[[1.0, -1.0], [1.0, 1.0]]])).all()
        assert grid.in_grouped_list == [[(1.0, -1.0), (1.0, 1.0)]]
        assert grid.in_list == [(1.0, -1.0), (1.0, 1.0)]

        # Input list

        grid = aa.Grid2DIrregularGrouped(grid=[[1.0, -1.0], [1.0, 1.0]])

        assert type(grid) == grids.Grid2DIrregularGrouped
        assert (grid == np.array([[[1.0, -1.0], [1.0, 1.0]]])).all()
        assert grid.in_grouped_list == [[(1.0, -1.0), (1.0, 1.0)]]
        assert grid.in_list == [(1.0, -1.0), (1.0, 1.0)]

        grid = aa.Grid2DIrregularGrouped(grid=[[[1.0, -1.0]], [[1.0, 1.0]]])

        assert type(grid) == grids.Grid2DIrregularGrouped
        assert (grid == np.array([[[1.0, -1.0], [1.0, 1.0]]])).all()
        assert grid.in_grouped_list == [[(1.0, -1.0)], [(1.0, 1.0)]]
        assert grid.in_list == [(1.0, -1.0), (1.0, 1.0)]

        grid = aa.Grid2DIrregularGrouped(grid=[[[1.0, -1.0], [1.0, 1.0]]])

        assert type(grid) == grids.Grid2DIrregularGrouped
        assert (grid == np.array([[[1.0, -1.0], [1.0, 1.0]]])).all()
        assert grid.in_grouped_list == [[(1.0, -1.0), (1.0, 1.0)]]
        assert grid.in_list == [(1.0, -1.0), (1.0, 1.0)]

    def test__input_as_dict__retains_dict_and_convert_correctly(self):

        # Input tuples

        grid = aa.Grid2DIrregularGrouped(
            grid=dict(source_0=[(1.0, -1.0), (1.0, 1.0)], source_1=[(2.0, 2.0)])
        )

        assert type(grid) == grids.Grid2DIrregularGrouped
        assert (grid == np.array([[[1.0, -1.0], [1.0, 1.0], [2.0, 2.0]]])).all()
        assert grid.in_grouped_list == [[(1.0, -1.0), (1.0, 1.0)], [(2.0, 2.0)]]
        assert grid.in_list == [(1.0, -1.0), (1.0, 1.0), (2.0, 2.0)]
        assert grid.as_dict["source_0"] == [(1.0, -1.0), (1.0, 1.0)]
        assert grid.as_dict["source_1"] == [(2.0, 2.0)]

    def test__from_yx_1d(self):
        grid = aa.Grid2DIrregularGrouped.from_yx_1d(y=[1.0, 1.0], x=[-1.0, 1.0])

        assert grid.in_grouped_list == [[(1.0, -1.0), (1.0, 1.0)]]

        grid = aa.Grid2DIrregularGrouped.from_yx_1d(
            y=[1.0, 1.0, 2.0, 4.0], x=[-1.0, 1.0, 3.0, 5.0]
        )

        assert grid.in_grouped_list == [
            [(1.0, -1.0), (1.0, 1.0), (2.0, 3.0), (4.0, 5.0)]
        ]

    def test__values_from_array_slim(self):

        grid = aa.Grid2DIrregularGrouped(grid=[[(1.0, 1.0), (2.0, 2.0)]])

        values_from_1d = grid.values_from_array_slim(array_slim=np.array([1.0, 2.0]))

        assert values_from_1d.in_grouped_list == [[1.0, 2.0]]

        grid = aa.Grid2DIrregularGrouped(grid=[[(1.0, 1.0), (2.0, 2.0)], [(3.0, 3.0)]])

        values_from_1d = grid.values_from_array_slim(
            array_slim=np.array([1.0, 2.0, 3.0])
        )

        assert values_from_1d.in_grouped_list == [[1.0, 2.0], [3.0]]

    def test__values_from_value(self):

        grid = aa.Grid2DIrregularGrouped(grid=[[(1.0, 1.0), (2.0, 2.0)]])

        values_from_value = grid.values_from_value(value=1.0)

        assert values_from_value.in_grouped_list == [[1.0, 1.0]]

        grid = aa.Grid2DIrregularGrouped(grid=[[(1.0, 1.0), (2.0, 2.0)], [(3.0, 3.0)]])

        values_from_value = grid.values_from_value(value=2.0)

        assert values_from_value.in_grouped_list == [[2.0, 2.0], [2.0]]

    def test__grid_from_grid_slim(self):

        grid = aa.Grid2DIrregularGrouped(grid=[[(1.0, 1.0), (2.0, 2.0)]])

        grid_from_1d = grid.grid_from_grid_slim(
            grid_slim=np.array([[1.0, 1.0], [2.0, 2.0]])
        )

        assert type(grid_from_1d) == grids.Grid2DIrregularGrouped
        assert grid_from_1d.in_grouped_list == [[(1.0, 1.0), (2.0, 2.0)]]

        grid = aa.Grid2DIrregularGrouped(grid=[[(1.0, 1.0), (2.0, 2.0)], [(3.0, 3.0)]])

        grid_from_1d = grid.grid_from_grid_slim(
            grid_slim=np.array([[1.0, 1.0], [2.0, 2.0], [3.0, 3.0]])
        )

        assert type(grid_from_1d) == grids.Grid2DIrregularGrouped
        assert grid_from_1d.in_grouped_list == [[(1.0, 1.0), (2.0, 2.0)], [(3.0, 3.0)]]

        grid_from_1d = grid.grid_from_grid_slim(
            grid_slim=np.array([[1.0, 1.0], [2.0, 2.0], [3.0, 3.0]])
        )

        assert type(grid_from_1d) == grids.Grid2DIrregularGrouped
        assert grid_from_1d.in_grouped_list == [[(1.0, 1.0), (2.0, 2.0)], [(3.0, 3.0)]]

    def test__grid_from_deflection_grid(self):
        grid = aa.Grid2DIrregularGrouped(grid=[[(1.0, 1.0), (2.0, 2.0)]])

        grid = grid.grid_from_deflection_grid(
            deflection_grid=np.array([[1.0, 0.0], [1.0, 1.0]])
        )

        assert grid.in_grouped_list == [[(0.0, 1.0), (1.0, 1.0)]]

    def test__furthest_distances_from_other_coordinates(self):

        grid = aa.Grid2DIrregularGrouped(grid=[[(0.0, 0.0), (0.0, 1.0)]])

        assert grid.furthest_distances_from_other_coordinates.in_grouped_list == [
            [1.0, 1.0]
        ]

        grid = aa.Grid2DIrregularGrouped(grid=[[(2.0, 4.0), (3.0, 6.0)]])

        assert grid.furthest_distances_from_other_coordinates.in_grouped_list == [
            [np.sqrt(5), np.sqrt(5)]
        ]

        grid = aa.Grid2DIrregularGrouped(grid=[[(0.0, 0.0), (0.0, 1.0), (0.0, 3.0)]])

        assert grid.furthest_distances_from_other_coordinates.in_grouped_list == [
            [3.0, 2.0, 3.0]
        ]

        grid = aa.Grid2DIrregularGrouped(
            grid=[[(0.0, 0.0), (0.0, 1.0), (0.0, 3.0)], [(0.0, 0.0)]]
        )

        assert grid.furthest_distances_from_other_coordinates.in_grouped_list == [
            [3.0, 2.0, 3.0],
            [0.0],
        ]

    def test__grid_of_closest_from_grid(self):

        grid = aa.Grid2DIrregularGrouped(grid=[[(0.0, 0.0), (0.0, 1.0)]])
        grid_pair = aa.Grid2DIrregularGrouped(grid=[[(0.0, 0.1)]])

        grid_of_closest = grid.grid_of_closest_from_grid_pair(grid_pair=grid_pair)

        assert grid_of_closest.in_grouped_list == [[(0.0, 0.0)]]

        grid_pair = aa.Grid2DIrregularGrouped(
            grid=[[(0.0, 0.1), (0.0, 0.2), (0.0, 0.3)]]
        )

        grid_of_closest = grid.grid_of_closest_from_grid_pair(grid_pair=grid_pair)

        assert grid_of_closest.in_grouped_list == [[(0.0, 0.0), (0.0, 0.0), (0.0, 0.0)]]

        grid_pair = aa.Grid2DIrregularGrouped(
            grid=[[(0.0, 0.1), (0.0, 0.2), (0.0, 0.9), (0.0, -0.1)]]
        )

        grid_of_closest = grid.grid_of_closest_from_grid_pair(grid_pair=grid_pair)

        assert grid_of_closest.in_grouped_list == [
            [(0.0, 0.0), (0.0, 0.0), (0.0, 1.0), (0.0, 0.0)]
        ]

        grid = aa.Grid2DIrregularGrouped(
            grid=[[(0.0, 0.0), (0.0, 1.0)], [(0.0, 0.0), (0.0, 1.0)]]
        )
        grid_pair = aa.Grid2DIrregularGrouped(
            grid=[
                [(0.0, 0.1), (0.0, 0.2), (0.0, 0.3)],
                [(0.0, 0.1), (0.0, 0.2), (0.0, 0.9), (0.0, -0.1)],
            ]
        )

        grid_of_closest = grid.grid_of_closest_from_grid_pair(grid_pair=grid_pair)

        assert grid_of_closest.in_grouped_list == [
            [(0.0, 0.0), (0.0, 0.0), (0.0, 0.0)],
            [(0.0, 0.0), (0.0, 0.0), (0.0, 1.0), (0.0, 0.0)],
        ]

    def test__with_mask__converts_to_and_from_pixels(self):

        mask = aa.Mask2D.manual(
            mask=np.full(fill_value=False, shape=(2, 2)), pixel_scales=(2.0, 2.0)
        )

        grid = aa.Grid2DIrregularGrouped(grid=[[(1.0, -1.0), (1.0, 1.0)]])

        assert grid.in_grouped_list == [[(1.0, -1.0), (1.0, 1.0)]]

        grid = aa.Grid2DIrregularGrouped.from_pixels_and_mask(
            pixels=[[(0, 0), (0, 1)]], mask=mask
        )

        assert grid.in_grouped_list == [[(1.0, -1.0), (1.0, 1.0)]]

    def test__load_grid__retains_list_structure(self):

        file_path = path.join(test_grid_dir, "output_test", "grid_test.dat")

        grid = aa.Grid2DIrregularGrouped.from_file(file_path=file_path)

        assert grid.in_grouped_list == [
            [(4.0, 4.0), (5.0, 5.0)],
            [(6.0, 6.0), (7.0, 7.0), (8.0, 8.0)],
        ]

    def test__output_grid_to_file(self):
        grid = aa.Grid2DIrregularGrouped(
            [[(4.0, 4.0), (5.0, 5.0)], [(6.0, 6.0), (7.0, 7.0), (8.0, 8.0)]]
        )

        output_grid_dir = path.join(
            "{}".format(os.path.dirname(os.path.realpath(__file__))),
            "files",
            "grid",
            "output_test",
        )

        file_path = path.join(output_grid_dir, "grid_test.dat")

        if os.path.exists(output_grid_dir):
            shutil.rmtree(output_grid_dir)

        os.makedirs(output_grid_dir)

        grid.output_to_file(file_path=file_path)

        grid = aa.Grid2DIrregularGrouped.from_file(file_path=file_path)

        assert grid.in_grouped_list == [
            [(4.0, 4.0), (5.0, 5.0)],
            [(6.0, 6.0), (7.0, 7.0), (8.0, 8.0)],
        ]

        with pytest.raises(FileExistsError):
            grid.output_to_file(file_path=file_path)

        grid.output_to_file(file_path=file_path, overwrite=True)


class TestGrid2DIrregularGroupedUniform:
    def test__from_sparse_uniform_upscale(self):

        grid_sparse_uniform = aa.Grid2DIrregularGroupedUniform(
            grid=[[(1.0, 1.0), (1.0, 3.0)]], pixel_scales=2.0
        )

        grid_upscale = grids.Grid2DIrregularGroupedUniform.from_grid_sparse_uniform_upscale(
            grid_sparse_uniform=grid_sparse_uniform, upscale_factor=2, pixel_scales=2.0
        )

        assert (
            grid_upscale
            == np.array(
                [
                    [1.5, 0.5],
                    [1.5, 1.5],
                    [0.5, 0.5],
                    [0.5, 1.5],
                    [1.5, 2.5],
                    [1.5, 3.5],
                    [0.5, 2.5],
                    [0.5, 3.5],
                ]
            )
        ).all()

        grid_sparse_uniform = aa.Grid2DIrregularGroupedUniform(
            grid=[[(1.0, 1.0), (1.0, 3.0), (1.0, 5.0), (3.0, 3.0)]], pixel_scales=2.0
        )

        grid_upscale = grids.Grid2DIrregularGroupedUniform.from_grid_sparse_uniform_upscale(
            grid_sparse_uniform=grid_sparse_uniform, upscale_factor=4, pixel_scales=2.0
        )

        grid_upscale_util = aa.util.grid.grid_2d_slim_upscaled_from(
            grid_slim=grid_sparse_uniform, upscale_factor=4, pixel_scales=(2.0, 2.0)
        )

        assert (grid_upscale == grid_upscale_util).all()

    def test__grid_from_deflection_grid(self):
        grid = aa.Grid2DIrregularGroupedUniform(
            grid=[[(1.0, 1.0), (2.0, 2.0)]],
            pixel_scales=(1.0, 1.0),
            shape_native=(3, 3),
        )

        grid = grid.grid_from_deflection_grid(
            deflection_grid=np.array([[1.0, 0.0], [1.0, 1.0]])
        )

        assert type(grid) == grids.Grid2DIrregularGroupedUniform
        assert grid.in_grouped_list == [[(0.0, 1.0), (1.0, 1.0)]]
        assert grid.pixel_scales == (1.0, 1.0)
        assert grid.shape_native == (3, 3)

    def test__grid_from_grid_slim(self):

        grid = aa.Grid2DIrregularGroupedUniform(
            grid=[[(1.0, 1.0), (2.0, 2.0)]],
            pixel_scales=(1.0, 1.0),
            shape_native=(3, 3),
        )

        grid_from_1d = grid.grid_from_grid_slim(
            grid_slim=np.array([[1.0, 1.0], [2.0, 2.0]])
        )

        assert type(grid_from_1d) == grids.Grid2DIrregularGroupedUniform
        assert grid_from_1d.in_grouped_list == [[(1.0, 1.0), (2.0, 2.0)]]
        assert grid.pixel_scales == (1.0, 1.0)
        assert grid.shape_native == (3, 3)

        grid = aa.Grid2DIrregularGroupedUniform(
            grid=[[(1.0, 1.0), (2.0, 2.0)], [(3.0, 3.0)]],
            pixel_scales=(1.0, 3.0),
            shape_native=(5, 5),
        )

        grid_from_1d = grid.grid_from_grid_slim(
            grid_slim=np.array([[1.0, 1.0], [2.0, 2.0], [3.0, 3.0]])
        )

        assert type(grid_from_1d) == grids.Grid2DIrregularGroupedUniform
        assert grid_from_1d.in_grouped_list == [[(1.0, 1.0), (2.0, 2.0)], [(3.0, 3.0)]]
        assert grid.pixel_scales == (1.0, 3.0)
        assert grid.shape_native == (5, 5)



import os

import numpy as np
import autoarray as aa


path = "{}/".format(os.path.dirname(os.path.realpath(__file__)))


class TestFrameAPI:
    def test__manual__makes_frame_using_inputs__include_rotations(self):

        frame = aa.Frame2D.manual(
            array=[[1.0, 2.0], [3.0, 4.0]],
            pixel_scales=1.0,
            roe_corner=(1, 0),
            scans=aa.Scans(
                parallel_overscan=(0, 1, 0, 1),
                serial_prescan=(1, 2, 1, 2),
                serial_overscan=(0, 2, 0, 2),
            ),
        )

        assert (frame == np.array([[1.0, 2.0], [3.0, 4.0]])).all()
        assert frame.original_roe_corner == (1, 0)
        assert frame.scans.parallel_overscan == (0, 1, 0, 1)
        assert frame.scans.serial_prescan == (1, 2, 1, 2)
        assert frame.scans.serial_overscan == (0, 2, 0, 2)
        assert (frame.mask == np.array([[False, False], [False, False]])).all()
        assert (frame.original_orientation == np.array([[1.0, 2.0], [3.0, 4.0]])).all()
        assert (frame.native == np.array([[1.0, 2.0], [3.0, 4.0]])).all()
        assert frame.native.scans.parallel_overscan == (0, 1, 0, 1)

        frame = aa.Frame2D.manual(
            array=[[1.0, 2.0], [3.0, 4.0]],
            pixel_scales=1.0,
            roe_corner=(0, 0),
            scans=aa.Scans(
                parallel_overscan=(0, 1, 0, 1),
                serial_prescan=(1, 2, 1, 2),
                serial_overscan=(0, 2, 0, 2),
            ),
        )

        assert (frame == np.array([[3.0, 4.0], [1.0, 2.0]])).all()
        assert frame.original_roe_corner == (0, 0)
        assert frame.scans.parallel_overscan == (1, 2, 0, 1)
        assert frame.scans.serial_prescan == (0, 1, 1, 2)
        assert frame.scans.serial_overscan == (0, 2, 0, 2)
        assert (frame.mask == np.array([[False, False], [False, False]])).all()
        assert (frame.original_orientation == np.array([[1.0, 2.0], [3.0, 4.0]])).all()

        frame = aa.Frame2D.manual(
            array=[[1.0, 2.0], [3.0, 4.0]],
            pixel_scales=1.0,
            roe_corner=(1, 1),
            scans=aa.Scans(
                parallel_overscan=(0, 1, 0, 1),
                serial_prescan=(1, 2, 1, 2),
                serial_overscan=(0, 2, 0, 2),
            ),
        )

        assert (frame == np.array([[2.0, 1.0], [4.0, 3.0]])).all()
        assert frame.original_roe_corner == (1, 1)
        assert frame.scans.parallel_overscan == (0, 1, 1, 2)
        assert frame.scans.serial_prescan == (1, 2, 0, 1)
        assert frame.scans.serial_overscan == (0, 2, 0, 2)
        assert (frame.mask == np.array([[False, False], [False, False]])).all()
        assert (frame.original_orientation == np.array([[1.0, 2.0], [3.0, 4.0]])).all()

        frame = aa.Frame2D.manual(
            array=[[1.0, 2.0], [3.0, 4.0]],
            pixel_scales=1.0,
            roe_corner=(0, 1),
            scans=aa.Scans(
                parallel_overscan=(0, 1, 0, 1),
                serial_prescan=(1, 2, 1, 2),
                serial_overscan=(0, 2, 0, 2),
            ),
        )

        assert (frame == np.array([[4.0, 3.0], [2.0, 1.0]])).all()
        assert frame.original_roe_corner == (0, 1)
        assert frame.scans.parallel_overscan == (1, 2, 1, 2)
        assert frame.scans.serial_prescan == (0, 1, 0, 1)
        assert frame.scans.serial_overscan == (0, 2, 0, 2)
        assert (frame.mask == np.array([[False, False], [False, False]])).all()
        assert (frame.original_orientation == np.array([[1.0, 2.0], [3.0, 4.0]])).all()

    def test__full_ones_zeros__makes_frame_using_inputs(self):

        frame = aa.Frame2D.full(
            fill_value=8.0,
            shape_native=(2, 2),
            pixel_scales=1.0,
            roe_corner=(1, 0),
            scans=aa.Scans(
                parallel_overscan=(0, 1, 0, 1),
                serial_prescan=(1, 2, 1, 2),
                serial_overscan=(0, 2, 0, 2),
            ),
        )

        assert (frame == np.array([[8.0, 8.0], [8.0, 8.0]])).all()
        assert frame.original_roe_corner == (1, 0)
        assert frame.scans.parallel_overscan == (0, 1, 0, 1)
        assert frame.scans.serial_prescan == (1, 2, 1, 2)
        assert frame.scans.serial_overscan == (0, 2, 0, 2)
        assert (frame.mask == np.array([[False, False], [False, False]])).all()

        frame = aa.Frame2D.ones(
            shape_native=(2, 2),
            pixel_scales=1.0,
            roe_corner=(1, 0),
            scans=aa.Scans(
                parallel_overscan=(0, 1, 0, 1),
                serial_prescan=(1, 2, 1, 2),
                serial_overscan=(0, 2, 0, 2),
            ),
        )

        assert (frame == np.array([[1.0, 1.0], [1.0, 1.0]])).all()
        assert frame.original_roe_corner == (1, 0)
        assert frame.scans.parallel_overscan == (0, 1, 0, 1)
        assert frame.scans.serial_prescan == (1, 2, 1, 2)
        assert frame.scans.serial_overscan == (0, 2, 0, 2)
        assert (frame.mask == np.array([[False, False], [False, False]])).all()

        frame = aa.Frame2D.zeros(
            shape_native=(2, 2),
            pixel_scales=1.0,
            roe_corner=(1, 0),
            scans=aa.Scans(
                parallel_overscan=(0, 1, 0, 1),
                serial_prescan=(1, 2, 1, 2),
                serial_overscan=(0, 2, 0, 2),
            ),
        )

        assert (frame == np.array([[0.0, 0.0], [0.0, 0.0]])).all()
        assert frame.original_roe_corner == (1, 0)
        assert frame.scans.parallel_overscan == (0, 1, 0, 1)
        assert frame.scans.serial_prescan == (1, 2, 1, 2)
        assert frame.scans.serial_overscan == (0, 2, 0, 2)
        assert (frame.mask == np.array([[False, False], [False, False]])).all()

    def test__extracted_frame_from_frame_and_extraction_region(self):

        frame = aa.Frame2D.manual(
            array=[[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]],
            pixel_scales=1.0,
            roe_corner=(1, 0),
            scans=aa.Scans(
                parallel_overscan=None,
                serial_prescan=(0, 2, 0, 2),
                serial_overscan=(1, 2, 1, 2),
            ),
        )

        frame = aa.Frame2D.extracted_frame_from_frame_and_extraction_region(
            frame=frame, extraction_region=aa.Region2D(region=(1, 3, 1, 3))
        )

        assert (frame == np.array([[5.0, 6.0], [8.0, 9.0]])).all()
        assert frame.original_roe_corner == (1, 0)
        assert frame.scans.parallel_overscan == None
        assert frame.scans.serial_prescan == (0, 1, 0, 1)
        assert frame.scans.serial_overscan == (0, 1, 0, 1)
        assert (frame.mask == np.array([[False, False], [False, False]])).all()

    def test__manual_mask__makes_frame_using_inputs__includes_rotation_which_includes_mask(
        self,
    ):

        mask = aa.Mask2D.manual(mask=[[False, True], [False, False]], pixel_scales=1.0)

        frame = aa.Frame2D.manual_mask(
            array=[[1.0, 2.0], [3.0, 4.0]],
            mask=mask,
            roe_corner=(1, 0),
            scans=aa.Scans(
                parallel_overscan=(0, 1, 0, 1),
                serial_prescan=(1, 2, 1, 2),
                serial_overscan=(0, 2, 0, 2),
            ),
        )

        assert (frame == np.array([[1.0, 0.0], [3.0, 4.0]])).all()
        assert frame.original_roe_corner == (1, 0)
        assert frame.scans.parallel_overscan == (0, 1, 0, 1)
        assert frame.scans.serial_prescan == (1, 2, 1, 2)
        assert frame.scans.serial_overscan == (0, 2, 0, 2)
        assert (frame.mask == np.array([[False, True], [False, False]])).all()

        frame = aa.Frame2D.manual_mask(
            array=[[1.0, 2.0], [3.0, 4.0]],
            mask=mask,
            roe_corner=(0, 0),
            scans=aa.Scans(
                parallel_overscan=(0, 1, 0, 1),
                serial_prescan=(1, 2, 1, 2),
                serial_overscan=(0, 2, 0, 2),
            ),
        )

        assert (frame == np.array([[3.0, 4.0], [1.0, 0.0]])).all()
        assert frame.original_roe_corner == (0, 0)
        assert frame.scans.parallel_overscan == (1, 2, 0, 1)
        assert frame.scans.serial_prescan == (0, 1, 1, 2)
        assert frame.scans.serial_overscan == (0, 2, 0, 2)
        assert (frame.mask == np.array([[False, False], [False, True]])).all()

        frame = aa.Frame2D.manual_mask(
            array=[[1.0, 2.0], [3.0, 4.0]],
            mask=mask,
            roe_corner=(1, 1),
            scans=aa.Scans(
                parallel_overscan=(0, 1, 0, 1),
                serial_prescan=(1, 2, 1, 2),
                serial_overscan=(0, 2, 0, 2),
            ),
        )

        assert (frame == np.array([[0.0, 1.0], [4.0, 3.0]])).all()
        assert frame.original_roe_corner == (1, 1)
        assert frame.scans.parallel_overscan == (0, 1, 1, 2)
        assert frame.scans.serial_prescan == (1, 2, 0, 1)
        assert frame.scans.serial_overscan == (0, 2, 0, 2)
        assert (frame.mask == np.array([[True, False], [False, False]])).all()

        frame = aa.Frame2D.manual_mask(
            array=[[1.0, 2.0], [3.0, 4.0]],
            mask=mask,
            roe_corner=(0, 1),
            scans=aa.Scans(
                parallel_overscan=(0, 1, 0, 1),
                serial_prescan=(1, 2, 1, 2),
                serial_overscan=(0, 2, 0, 2),
            ),
        )

        assert (frame == np.array([[4.0, 3.0], [0.0, 1.0]])).all()
        assert frame.original_roe_corner == (0, 1)
        assert frame.scans.parallel_overscan == (1, 2, 1, 2)
        assert frame.scans.serial_prescan == (0, 1, 0, 1)
        assert frame.scans.serial_overscan == (0, 2, 0, 2)
        assert (frame.mask == np.array([[False, False], [True, False]])).all()

    def test__from_frame__no_rotation_as_frame_is_correct_orientation(self):

        mask = aa.Mask2D.manual(mask=[[False, True], [False, False]], pixel_scales=1.0)

        frame = aa.Frame2D.manual(
            array=[[1.0, 2.0], [3.0, 4.0]],
            pixel_scales=1.0,
            roe_corner=(1, 0),
            scans=aa.Scans(
                parallel_overscan=(0, 1, 0, 1),
                serial_prescan=(1, 2, 1, 2),
                serial_overscan=(0, 2, 0, 2),
            ),
        )

        frame = aa.Frame2D.from_frame(frame=frame, mask=mask)

        assert (frame == np.array([[1.0, 0.0], [3.0, 4.0]])).all()
        assert frame.original_roe_corner == (1, 0)
        assert frame.scans.parallel_overscan == (0, 1, 0, 1)
        assert frame.scans.serial_prescan == (1, 2, 1, 2)
        assert frame.scans.serial_overscan == (0, 2, 0, 2)
        assert (frame.mask == np.array([[False, True], [False, False]])).all()

        mask = aa.Mask2D.manual(mask=[[False, True], [False, False]], pixel_scales=1.0)

        frame = aa.Frame2D.manual(
            array=[[1.0, 2.0], [3.0, 4.0]],
            pixel_scales=1.0,
            roe_corner=(0, 0),
            scans=aa.Scans(
                parallel_overscan=(0, 1, 0, 1),
                serial_prescan=(1, 2, 1, 2),
                serial_overscan=(0, 2, 0, 2),
            ),
        )

        frame = aa.Frame2D.from_frame(frame=frame, mask=mask)

        assert (frame == np.array([[3.0, 0.0], [1.0, 2.0]])).all()
        assert frame.original_roe_corner == (0, 0)
        assert frame.scans.parallel_overscan == (1, 2, 0, 1)
        assert frame.scans.serial_prescan == (0, 1, 1, 2)
        assert frame.scans.serial_overscan == (0, 2, 0, 2)
        assert (frame.mask == np.array([[False, True], [False, False]])).all()



from os import path
import matplotlib.pyplot as plt
import autoarray.plot as aplt
from autoarray.plot.plotters import abstract_plotters

directory = path.dirname(path.realpath(__file__))


class TestAbstractPlotter:
    def test__subplot_figsize_for_number_of_subplots(self):

        plotter = abstract_plotters.AbstractPlotter()

        figsize = plotter.get_subplot_figsize(number_subplots=1)

        assert figsize == (18, 8)

        figsize = plotter.get_subplot_figsize(number_subplots=4)

        assert figsize == (13, 10)

        figure = aplt.Figure(figsize=(20, 20))

        plotter = abstract_plotters.AbstractPlotter(
            mat_plot_2d=aplt.MatPlot2D(figure=figure)
        )

        figsize = plotter.get_subplot_figsize(number_subplots=4)

        assert figsize == (20, 20)

    def test__plotter_number_of_subplots(self):

        plotter = abstract_plotters.AbstractPlotter(mat_plot_2d=aplt.MatPlot2D())

        rows, columns = plotter.mat_plot_2d.get_subplot_rows_columns(number_subplots=1)

        assert rows == 1
        assert columns == 2

        rows, columns = plotter.mat_plot_2d.get_subplot_rows_columns(number_subplots=4)

        assert rows == 2
        assert columns == 2

    def test__open_and_close_subplot_figures(self):

        figure = aplt.Figure(figsize=(20, 20))

        plotter = abstract_plotters.AbstractPlotter(
            mat_plot_2d=aplt.MatPlot2D(figure=figure)
        )

        plotter.mat_plot_2d.figure.open()

        assert plt.fignum_exists(num=1) == True

        plotter.mat_plot_2d.figure.close()

        assert plt.fignum_exists(num=1) == False

        plotter = abstract_plotters.AbstractPlotter(
            mat_plot_2d=aplt.MatPlot2D(figure=figure)
        )

        assert plt.fignum_exists(num=1) == False

        plotter.open_subplot_figure(number_subplots=4)

        assert plt.fignum_exists(num=1) == True

        plotter.mat_plot_2d.figure.close()

        assert plt.fignum_exists(num=1) == False

    def test__uses_figure_or_subplot_configs_correctly(self):

        figure = aplt.Figure(figsize=(8, 8))
        cmap = aplt.Cmap(cmap="warm")

        mat_plot_2d = aplt.MatPlot2D(figure=figure, cmap=cmap)

        plotter = abstract_plotters.AbstractPlotter(mat_plot_2d=mat_plot_2d)

        assert plotter.mat_plot_2d.figure.config_dict["figsize"] == (8, 8)
        assert plotter.mat_plot_2d.figure.config_dict["aspect"] == "square"
        assert plotter.mat_plot_2d.cmap.config_dict["cmap"] == "warm"
        assert plotter.mat_plot_2d.cmap.config_dict["norm"] == "linear"

        figure = aplt.Figure()
        figure.is_for_subplot = True

        cmap = aplt.Cmap()
        cmap.is_for_subplot = True

        mat_plot_2d = aplt.MatPlot2D(figure=figure, cmap=cmap)

        plotter = abstract_plotters.AbstractPlotter(mat_plot_2d=mat_plot_2d)

        assert plotter.mat_plot_2d.figure.config_dict["figsize"] == None
        assert plotter.mat_plot_2d.figure.config_dict["aspect"] == "square"
        assert plotter.mat_plot_2d.cmap.config_dict["cmap"] == "jet"
        assert plotter.mat_plot_2d.cmap.config_dict["norm"] == "linear"

    def test__attribute_for_visuals(self):

        visuals_2d = aplt.Visuals2D()
        include_2d = aplt.Include2D(origin=False)

        abstract_plotter = abstract_plotters.AbstractPlotter(
            visuals_2d=visuals_2d, include_2d=include_2d
        )
        attr = abstract_plotter.extract_2d(name="origin", value=1)

        assert attr == None

        include_2d = aplt.Include2D(origin=True)
        abstract_plotter = abstract_plotters.AbstractPlotter(
            visuals_2d=visuals_2d, include_2d=include_2d
        )
        attr = abstract_plotter.extract_2d(name="origin", value=1)

        assert attr == 1

        visuals_2d = aplt.Visuals2D(origin=10)

        include_2d = aplt.Include2D(origin=False)
        abstract_plotter = abstract_plotters.AbstractPlotter(
            visuals_2d=visuals_2d, include_2d=include_2d
        )
        attr = abstract_plotter.extract_2d(name="origin", value=2)

        assert attr == 10

        include_2d = aplt.Include2D(origin=True)
        abstract_plotter = abstract_plotters.AbstractPlotter(
            visuals_2d=visuals_2d, include_2d=include_2d
        )
        attr = abstract_plotter.extract_2d(name="origin", value=2)

        assert attr == 10



import datetime

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

# -- Project information -----------------------------------------------------

year = datetime.date.today().year
project = "PyAutoArray"
copyright = "2020, James Nightingale, Richard Hayes"
author = "James Nightingale, Richard Hayes"

# The full version, including alpha/beta/rc tags
release = "1.0.16"
master_doc = "index"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.extlinks",
    "numpydoc",
    "nbsphinx",
]

## Generate autodoc stubs with summaries from code
autosummary_generate = True

## Include Python objects as they appear in source files
autodoc_member_order = "bysource"

## Default flags used by autodoc directives
autodoc_default_flags = ["members"]

numpydoc_show_class_members = False
numpydoc_show_inherited_class_members = False
numpydoc_class_members_toctree = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


sphinx_gallery_conf = {
    # Remove the "Download all examples" button from the top level gallery
    "download_all_examples": False,
    # directory where function granular galleries are stored
    "backreferences_dir": "api/generated/backreferences",
    # Modules for which function level galleries are created.
    "doc_module": "pyautoarray",
    # Insert links to documentation of objects in the examples
    "reference_url": {"pyautoarray": None},
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"
html_last_updated_fmt = "%b %d, %Y"
html_title = "PyAutoArray"
html_short_title = "PyAutoArray"
pygments_style = "default"
add_function_parentheses = False
html_show_sourcelink = False
html_show_sphinx = True
html_show_copyright = True

html_context = {
    "menu_links_name": "Repository",
    # Custom variables to enable "Improve this page"" and "Download notebook"
    # links
    "doc_path": "docs",
    "github_project": "pyautoarray",
    "github_repo": "pyautoarray",
    "github_version": "development",
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]



import autoarray as aa


class TestCoordinates1D:
    def test__central_pixel__depends_on_shape_pixel_scale_and_origin(self):

        central_pixel_coordinates = aa.util.geometry.central_pixel_coordinates_1d_from(
            shape_slim=(3,)
        )

        assert central_pixel_coordinates == (1,)

        central_pixel_coordinates = aa.util.geometry.central_pixel_coordinates_1d_from(
            shape_slim=(4,)
        )

        assert central_pixel_coordinates == (1.5,)

    def test__pixel_coordinates_1d_from(self):

        pixel_coordinates = aa.util.geometry.pixel_coordinates_1d_from(
            scaled_coordinates_1d=(1.0,), shape_slim=(2,), pixel_scales=(2.0,)
        )

        assert pixel_coordinates == (1,)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_1d_from(
            scaled_coordinates_1d=(1.0,), shape_slim=(2,), pixel_scales=(2.0,)
        )

        assert pixel_coordinates == (1,)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_1d_from(
            scaled_coordinates_1d=(-1.0,), shape_slim=(2,), pixel_scales=(2.0,)
        )

        assert pixel_coordinates == (0,)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_1d_from(
            scaled_coordinates_1d=(0.0,), shape_slim=(3,), pixel_scales=(3.0,)
        )

        assert pixel_coordinates == (1,)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_1d_from(
            scaled_coordinates_1d=(3.0,), shape_slim=(3,), pixel_scales=(3.0,)
        )

        assert pixel_coordinates == (2,)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_1d_from(
            scaled_coordinates_1d=(-3.0,), shape_slim=(3,), pixel_scales=(3.0,)
        )

        assert pixel_coordinates == (0,)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_1d_from(
            scaled_coordinates_1d=(3.0,), shape_slim=(3,), pixel_scales=(3.0,)
        )

        assert pixel_coordinates == (2,)

    def test__pixel_coordinates_1d_from__scaled_are_pixel_corners(self):

        pixel_coordinates = aa.util.geometry.pixel_coordinates_1d_from(
            scaled_coordinates_1d=(-1.99,), shape_slim=(2,), pixel_scales=(2.0,)
        )

        assert pixel_coordinates == (0,)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_1d_from(
            scaled_coordinates_1d=(-0.01,), shape_slim=(2,), pixel_scales=(2.0,)
        )

        assert pixel_coordinates == (0,)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_1d_from(
            scaled_coordinates_1d=(0.01,), shape_slim=(2,), pixel_scales=(2.0,)
        )

        assert pixel_coordinates == (1,)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_1d_from(
            scaled_coordinates_1d=(-1.99,), shape_slim=(2,), pixel_scales=(2.0,)
        )

        assert pixel_coordinates == (0,)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_1d_from(
            scaled_coordinates_1d=(-0.01,), shape_slim=(2,), pixel_scales=(2.0,)
        )

        assert pixel_coordinates == (0,)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_1d_from(
            scaled_coordinates_1d=(0.01,), shape_slim=(2,), pixel_scales=(2.0,)
        )

        assert pixel_coordinates == (1,)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_1d_from(
            scaled_coordinates_1d=(1.99,), shape_slim=(2,), pixel_scales=(2.0,)
        )

        assert pixel_coordinates == (1,)

    def test__pixel_coordinates_1d_from___scaled_are_pixel_centres__nonzero_centre(
        self,
    ):
        pixel_coordinates = aa.util.geometry.pixel_coordinates_1d_from(
            scaled_coordinates_1d=(0.0,),
            shape_slim=(2,),
            pixel_scales=(2.0,),
            origins=(1.0,),
        )

        assert pixel_coordinates == (0,)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_1d_from(
            scaled_coordinates_1d=(2.0,),
            shape_slim=(2,),
            pixel_scales=(2.0,),
            origins=(1.0,),
        )

        assert pixel_coordinates == (1,)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_1d_from(
            scaled_coordinates_1d=(0.0,),
            shape_slim=(3,),
            pixel_scales=(3.0,),
            origins=(3.0,),
        )

        assert pixel_coordinates == (0,)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_1d_from(
            scaled_coordinates_1d=(3.0,),
            shape_slim=(3,),
            pixel_scales=(3.0,),
            origins=(3.0,),
        )

        assert pixel_coordinates == (1,)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_1d_from(
            scaled_coordinates_1d=(6.0,),
            shape_slim=(3,),
            pixel_scales=(3.0,),
            origins=(3.0,),
        )

        assert pixel_coordinates == (2,)

    def test__pixel_coordinates_1d_from__scaled_are_pixel_corners__nonzero_centre(
        self,
    ):
        pixel_coordinates = aa.util.geometry.pixel_coordinates_1d_from(
            scaled_coordinates_1d=(-0.99,),
            shape_slim=(2,),
            pixel_scales=(2.0,),
            origins=(1.0,),
        )

        assert pixel_coordinates == (0,)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_1d_from(
            scaled_coordinates_1d=(0.99,),
            shape_slim=(2,),
            pixel_scales=(2.0,),
            origins=(1.0,),
        )

        assert pixel_coordinates == (0,)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_1d_from(
            scaled_coordinates_1d=(1.01,),
            shape_slim=(2,),
            pixel_scales=(2.0,),
            origins=(1.0,),
        )

        assert pixel_coordinates == (1,)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_1d_from(
            scaled_coordinates_1d=(2.99,),
            shape_slim=(2,),
            pixel_scales=(2.0,),
            origins=(1.0,),
        )

        assert pixel_coordinates == (1,)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_1d_from(
            scaled_coordinates_1d=(1.01,),
            shape_slim=(2,),
            pixel_scales=(2.0,),
            origins=(1.0,),
        )

        assert pixel_coordinates == (1,)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_1d_from(
            scaled_coordinates_1d=(2.99,),
            shape_slim=(2,),
            pixel_scales=(2.0,),
            origins=(1.0,),
        )

        assert pixel_coordinates == (1,)

    def test__scaled_coordinates_1d_from___scaled_are_pixel_centres__nonzero_centre(
        self,
    ):
        scaled_coordinates = aa.util.geometry.scaled_coordinates_1d_from(
            pixel_coordinates_1d=(0,), shape_slim=(3,), pixel_scales=(3.0,)
        )

        assert scaled_coordinates == (-3.0,)

        scaled_coordinates = aa.util.geometry.scaled_coordinates_1d_from(
            pixel_coordinates_1d=(1,), shape_slim=(3,), pixel_scales=(3.0,)
        )

        assert scaled_coordinates == (0.0,)

        scaled_coordinates = aa.util.geometry.scaled_coordinates_1d_from(
            pixel_coordinates_1d=(2,), shape_slim=(3,), pixel_scales=(3.0,)
        )

        assert scaled_coordinates == (3.0,)

        scaled_coordinates = aa.util.geometry.scaled_coordinates_1d_from(
            pixel_coordinates_1d=(0,),
            shape_slim=(2,),
            pixel_scales=(2.0,),
            origins=(1.0,),
        )

        assert scaled_coordinates == (0.0,)

        scaled_coordinates = aa.util.geometry.scaled_coordinates_1d_from(
            pixel_coordinates_1d=(1,),
            shape_slim=(2,),
            pixel_scales=(2.0,),
            origins=(1.0,),
        )

        assert scaled_coordinates == (2.0,)

        scaled_coordinates = aa.util.geometry.scaled_coordinates_1d_from(
            pixel_coordinates_1d=(0,),
            shape_slim=(3,),
            pixel_scales=(3.0,),
            origins=(3.0,),
        )

        assert scaled_coordinates == (0.0,)

        scaled_coordinates = aa.util.geometry.scaled_coordinates_1d_from(
            pixel_coordinates_1d=(1,),
            shape_slim=(3,),
            pixel_scales=(3.0,),
            origins=(3.0,),
        )

        assert scaled_coordinates == (3.0,)

        scaled_coordinates = aa.util.geometry.scaled_coordinates_1d_from(
            pixel_coordinates_1d=(2,),
            shape_slim=(3,),
            pixel_scales=(3.0,),
            origins=(3.0,),
        )

        assert scaled_coordinates == (6.0,)


class TestCoordinates2D:
    def test__central_pixel__depends_on_shape_pixel_scale_and_origin(self):

        central_pixel_coordinates = aa.util.geometry.central_pixel_coordinates_2d_from(
            shape_native=(3, 3)
        )

        assert central_pixel_coordinates == (1, 1)

        central_pixel_coordinates = aa.util.geometry.central_pixel_coordinates_2d_from(
            shape_native=(3, 3)
        )

        assert central_pixel_coordinates == (1, 1)

        central_pixel_coordinates = aa.util.geometry.central_pixel_coordinates_2d_from(
            shape_native=(4, 4)
        )

        assert central_pixel_coordinates == (1.5, 1.5)

        central_pixel_coordinates = aa.util.geometry.central_pixel_coordinates_2d_from(
            shape_native=(4, 4)
        )

        assert central_pixel_coordinates == (1.5, 1.5)

    def test__pixel_coordinates_2d_from(self):

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(1.0, -1.0),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
        )

        assert pixel_coordinates == (0, 0)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(1.0, 1.0),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
        )

        assert pixel_coordinates == (0, 1)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(-1.0, -1.0),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
        )

        assert pixel_coordinates == (1, 0)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(-1.0, 1.0),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
        )

        assert pixel_coordinates == (1, 1)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(3.0, -3.0),
            shape_native=(3, 3),
            pixel_scales=(3.0, 3.0),
        )

        assert pixel_coordinates == (0, 0)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(3.0, 0.0),
            shape_native=(3, 3),
            pixel_scales=(3.0, 3.0),
        )

        assert pixel_coordinates == (0, 1)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(3.0, 3.0),
            shape_native=(3, 3),
            pixel_scales=(3.0, 3.0),
        )

        assert pixel_coordinates == (0, 2)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(0.0, -3.0),
            shape_native=(3, 3),
            pixel_scales=(3.0, 3.0),
        )

        assert pixel_coordinates == (1, 0)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(0.0, 0.0),
            shape_native=(3, 3),
            pixel_scales=(3.0, 3.0),
        )

        assert pixel_coordinates == (1, 1)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(0.0, 3.0),
            shape_native=(3, 3),
            pixel_scales=(3.0, 3.0),
        )

        assert pixel_coordinates == (1, 2)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(-3.0, -3.0),
            shape_native=(3, 3),
            pixel_scales=(3.0, 3.0),
        )

        assert pixel_coordinates == (2, 0)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(-3.0, 0.0),
            shape_native=(3, 3),
            pixel_scales=(3.0, 3.0),
        )

        assert pixel_coordinates == (2, 1)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(-3.0, 3.0),
            shape_native=(3, 3),
            pixel_scales=(3.0, 3.0),
        )

        assert pixel_coordinates == (2, 2)

    def test__pixel_coordinates_2d_from__scaled_are_pixel_corners(self):

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(1.99, -1.99),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
        )

        assert pixel_coordinates == (0, 0)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(1.99, -0.01),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
        )

        assert pixel_coordinates == (0, 0)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(0.01, -1.99),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
        )

        assert pixel_coordinates == (0, 0)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(0.01, -0.01),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
        )

        assert pixel_coordinates == (0, 0)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(2.01, 0.01),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
        )

        assert pixel_coordinates == (0, 1)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(2.01, 1.99),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
        )

        assert pixel_coordinates == (0, 1)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(0.01, 0.01),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
        )

        assert pixel_coordinates == (0, 1)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(0.01, 1.99),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
        )

        assert pixel_coordinates == (0, 1)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(-0.01, -1.99),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
        )

        assert pixel_coordinates == (1, 0)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(-0.01, -0.01),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
        )

        assert pixel_coordinates == (1, 0)
        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(-1.99, -1.99),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
        )

        assert pixel_coordinates == (1, 0)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(-1.99, -0.01),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
        )

        assert pixel_coordinates == (1, 0)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(-0.01, 0.01),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
        )

        assert pixel_coordinates == (1, 1)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(-0.01, 1.99),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
        )

        assert pixel_coordinates == (1, 1)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(-1.99, 0.01),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
        )

        assert pixel_coordinates == (1, 1)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(-1.99, 1.99),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
        )

        assert pixel_coordinates == (1, 1)

    def test__pixel_coordinates_2d_from___scaled_are_pixel_centres__nonzero_centre(
        self,
    ):

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(2.0, 0.0),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
            origins=(1.0, 1.0),
        )

        assert pixel_coordinates == (0, 0)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(2.0, 2.0),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
            origins=(1.0, 1.0),
        )

        assert pixel_coordinates == (0, 1)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(0.0, 0.0),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
            origins=(1.0, 1.0),
        )

        assert pixel_coordinates == (1, 0)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(0.0, 2.0),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
            origins=(1.0, 1.0),
        )

        assert pixel_coordinates == (1, 1)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(6.0, 0.0),
            shape_native=(3, 3),
            pixel_scales=(3.0, 3.0),
            origins=(3.0, 3.0),
        )

        assert pixel_coordinates == (0, 0)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(6.0, 3.0),
            shape_native=(3, 3),
            pixel_scales=(3.0, 3.0),
            origins=(3.0, 3.0),
        )

        assert pixel_coordinates == (0, 1)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(6.0, 6.0),
            shape_native=(3, 3),
            pixel_scales=(3.0, 3.0),
            origins=(3.0, 3.0),
        )

        assert pixel_coordinates == (0, 2)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(3.0, 0.0),
            shape_native=(3, 3),
            pixel_scales=(3.0, 3.0),
            origins=(3.0, 3.0),
        )

        assert pixel_coordinates == (1, 0)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(3.0, 3.0),
            shape_native=(3, 3),
            pixel_scales=(3.0, 3.0),
            origins=(3.0, 3.0),
        )

        assert pixel_coordinates == (1, 1)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(3.0, 6.0),
            shape_native=(3, 3),
            pixel_scales=(3.0, 3.0),
            origins=(3.0, 3.0),
        )

        assert pixel_coordinates == (1, 2)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(0.0, 0.0),
            shape_native=(3, 3),
            pixel_scales=(3.0, 3.0),
            origins=(3.0, 3.0),
        )

        assert pixel_coordinates == (2, 0)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(0.0, 3.0),
            shape_native=(3, 3),
            pixel_scales=(3.0, 3.0),
            origins=(3.0, 3.0),
        )

        assert pixel_coordinates == (2, 1)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(0.0, 6.0),
            shape_native=(3, 3),
            pixel_scales=(3.0, 3.0),
            origins=(3.0, 3.0),
        )

        assert pixel_coordinates == (2, 2)

    def test__pixel_coordinates_2d_from__scaled_are_pixel_corners__nonzero_centre(
        self,
    ):

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(2.99, -0.99),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
            origins=(1.0, 1.0),
        )

        assert pixel_coordinates == (0, 0)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(2.99, 0.99),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
            origins=(1.0, 1.0),
        )

        assert pixel_coordinates == (0, 0)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(1.01, -0.99),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
            origins=(1.0, 1.0),
        )

        assert pixel_coordinates == (0, 0)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(1.01, 0.99),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
            origins=(1.0, 1.0),
        )

        assert pixel_coordinates == (0, 0)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(3.01, 1.01),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
            origins=(1.0, 1.0),
        )

        assert pixel_coordinates == (0, 1)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(3.01, 2.99),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
            origins=(1.0, 1.0),
        )

        assert pixel_coordinates == (0, 1)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(1.01, 1.01),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
            origins=(1.0, 1.0),
        )

        assert pixel_coordinates == (0, 1)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(1.01, 2.99),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
            origins=(1.0, 1.0),
        )

        assert pixel_coordinates == (0, 1)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(0.99, -0.99),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
            origins=(1.0, 1.0),
        )

        assert pixel_coordinates == (1, 0)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(0.99, 0.99),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
            origins=(1.0, 1.0),
        )

        assert pixel_coordinates == (1, 0)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(-0.99, -0.99),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
            origins=(1.0, 1.0),
        )

        assert pixel_coordinates == (1, 0)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(-0.99, 0.99),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
            origins=(1.0, 1.0),
        )

        assert pixel_coordinates == (1, 0)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(0.99, 1.01),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
            origins=(1.0, 1.0),
        )

        assert pixel_coordinates == (1, 1)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(0.99, 2.99),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
            origins=(1.0, 1.0),
        )

        assert pixel_coordinates == (1, 1)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(-0.99, 1.01),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
            origins=(1.0, 1.0),
        )

        assert pixel_coordinates == (1, 1)

        pixel_coordinates = aa.util.geometry.pixel_coordinates_2d_from(
            scaled_coordinates_2d=(-0.99, 2.99),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
            origins=(1.0, 1.0),
        )

        assert pixel_coordinates == (1, 1)

    def test__scaled_coordinates_2d_from___scaled_are_pixel_centres__nonzero_centre(
        self,
    ):

        scaled_coordinates = aa.util.geometry.scaled_coordinates_2d_from(
            pixel_coordinates_2d=(0, 0), shape_native=(3, 3), pixel_scales=(3.0, 3.0)
        )

        assert scaled_coordinates == (3.0, -3.0)

        scaled_coordinates = aa.util.geometry.scaled_coordinates_2d_from(
            pixel_coordinates_2d=(0, 1), shape_native=(3, 3), pixel_scales=(3.0, 3.0)
        )

        assert scaled_coordinates == (3.0, 0.0)

        scaled_coordinates = aa.util.geometry.scaled_coordinates_2d_from(
            pixel_coordinates_2d=(0, 2), shape_native=(3, 3), pixel_scales=(3.0, 3.0)
        )

        assert scaled_coordinates == (3.0, 3.0)

        scaled_coordinates = aa.util.geometry.scaled_coordinates_2d_from(
            pixel_coordinates_2d=(1, 0), shape_native=(3, 3), pixel_scales=(3.0, 3.0)
        )

        assert scaled_coordinates == (0.0, -3.0)

        scaled_coordinates = aa.util.geometry.scaled_coordinates_2d_from(
            pixel_coordinates_2d=(1, 1), shape_native=(3, 3), pixel_scales=(3.0, 3.0)
        )

        assert scaled_coordinates == (0.0, 0.0)

        scaled_coordinates = aa.util.geometry.scaled_coordinates_2d_from(
            pixel_coordinates_2d=(1, 2), shape_native=(3, 3), pixel_scales=(3.0, 3.0)
        )

        assert scaled_coordinates == (0.0, 3.0)

        scaled_coordinates = aa.util.geometry.scaled_coordinates_2d_from(
            pixel_coordinates_2d=(2, 0), shape_native=(3, 3), pixel_scales=(3.0, 3.0)
        )

        assert scaled_coordinates == (-3.0, -3.0)

        scaled_coordinates = aa.util.geometry.scaled_coordinates_2d_from(
            pixel_coordinates_2d=(2, 1), shape_native=(3, 3), pixel_scales=(3.0, 3.0)
        )

        assert scaled_coordinates == (-3.0, 0.0)

        scaled_coordinates = aa.util.geometry.scaled_coordinates_2d_from(
            pixel_coordinates_2d=(2, 2), shape_native=(3, 3), pixel_scales=(3.0, 3.0)
        )

        assert scaled_coordinates == (-3.0, 3.0)

        scaled_coordinates = aa.util.geometry.scaled_coordinates_2d_from(
            pixel_coordinates_2d=(0, 0),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
            origins=(1.0, 1.0),
        )

        assert scaled_coordinates == (2.0, 0.0)

        scaled_coordinates = aa.util.geometry.scaled_coordinates_2d_from(
            pixel_coordinates_2d=(0, 1),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
            origins=(1.0, 1.0),
        )

        assert scaled_coordinates == (2.0, 2.0)

        scaled_coordinates = aa.util.geometry.scaled_coordinates_2d_from(
            pixel_coordinates_2d=(1, 0),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
            origins=(1.0, 1.0),
        )

        assert scaled_coordinates == (0.0, 0.0)

        scaled_coordinates = aa.util.geometry.scaled_coordinates_2d_from(
            pixel_coordinates_2d=(1, 1),
            shape_native=(2, 2),
            pixel_scales=(2.0, 2.0),
            origins=(1.0, 1.0),
        )

        assert scaled_coordinates == (0.0, 2.0)

        scaled_coordinates = aa.util.geometry.scaled_coordinates_2d_from(
            pixel_coordinates_2d=(0, 0),
            shape_native=(3, 3),
            pixel_scales=(3.0, 3.0),
            origins=(3.0, 3.0),
        )

        assert scaled_coordinates == (6.0, 0.0)

        scaled_coordinates = aa.util.geometry.scaled_coordinates_2d_from(
            pixel_coordinates_2d=(0, 1),
            shape_native=(3, 3),
            pixel_scales=(3.0, 3.0),
            origins=(3.0, 3.0),
        )

        assert scaled_coordinates == (6.0, 3.0)

        scaled_coordinates = aa.util.geometry.scaled_coordinates_2d_from(
            pixel_coordinates_2d=(0, 2),
            shape_native=(3, 3),
            pixel_scales=(3.0, 3.0),
            origins=(3.0, 3.0),
        )

        assert scaled_coordinates == (6.0, 6.0)

        scaled_coordinates = aa.util.geometry.scaled_coordinates_2d_from(
            pixel_coordinates_2d=(1, 0),
            shape_native=(3, 3),
            pixel_scales=(3.0, 3.0),
            origins=(3.0, 3.0),
        )

        assert scaled_coordinates == (3.0, 0.0)

        scaled_coordinates = aa.util.geometry.scaled_coordinates_2d_from(
            pixel_coordinates_2d=(1, 1),
            shape_native=(3, 3),
            pixel_scales=(3.0, 3.0),
            origins=(3.0, 3.0),
        )

        assert scaled_coordinates == (3.0, 3.0)

        scaled_coordinates = aa.util.geometry.scaled_coordinates_2d_from(
            pixel_coordinates_2d=(1, 2),
            shape_native=(3, 3),
            pixel_scales=(3.0, 3.0),
            origins=(3.0, 3.0),
        )

        assert scaled_coordinates == (3.0, 6.0)

        scaled_coordinates = aa.util.geometry.scaled_coordinates_2d_from(
            pixel_coordinates_2d=(2, 0),
            shape_native=(3, 3),
            pixel_scales=(3.0, 3.0),
            origins=(3.0, 3.0),
        )

        assert scaled_coordinates == (0.0, 0.0)

        scaled_coordinates = aa.util.geometry.scaled_coordinates_2d_from(
            pixel_coordinates_2d=(2, 1),
            shape_native=(3, 3),
            pixel_scales=(3.0, 3.0),
            origins=(3.0, 3.0),
        )

        assert scaled_coordinates == (0.0, 3.0)

        scaled_coordinates = aa.util.geometry.scaled_coordinates_2d_from(
            pixel_coordinates_2d=(2, 2),
            shape_native=(3, 3),
            pixel_scales=(3.0, 3.0),
            origins=(3.0, 3.0),
        )

        assert scaled_coordinates == (0.0, 6.0)



import autoarray as aa
import autoarray.plot as aplt
import numpy as np

grid = aa.Grid2D.uniform(shape_native=(11, 11), pixel_scales=1.0)

aplt.Grid2D(grid=grid, indexes=[0, 1, 2, 14])

aplt.Grid2D(grid=grid, indexes=[(0, 0), (2, 3), (4, 1), (8, 5)])

aplt.Grid2D(grid=grid, indexes=[[(0, 0), (2, 3)], [(4, 1), (8, 5)]])

aplt.Grid2D(grid=grid, indexes=[[[0, 0], [2, 3]], [[4, 1], [8, 5]]])
